from distances import Indel, DiscountedLevenshtein, CormodeLZ, Tichy, IterativeSubString, BaulieuXIII, Clement, DiceAsymmetricI, KuhnsIII, Overlap, PearsonII, WeightedJaccard, WarrensIV, Bag, RougeL, RatcliffObershelp, NCDbz2, FuzzyWuzzyPartialString, FuzzyWuzzyTokenSort, FuzzyWuzzyTokenSet, Editex, Typo,LIG3, SSK, Levenshtein, DoubleMetaphone, RefinedSoundex, PhoneticDistance
from collections import defaultdict

def make_distance_metrics(indel=False,
                          discounted_levenshtein=False,
                          tichy=False,
                          cormodel_z=False,
                          iterative_sub_string=False,
                          baulieu_xiii=False,
                          clement=False,
                          dice_asymmetrici=False,
                          kuhns_iii=False,
                          overlap=False,
                          pearson_ii=False,
                          weighted_jaccard=False,
                          warrens_iv=False,
                          bag=False,
                          rouge_l=False,
                          ratcliff_obershelp=False,
                          ncd_bz2=False,
                          fuzzy_wuzzy_partial_string=False,
                          fuzzy_wuzzy_token_sort=False,
                          fuzzy_wuzzy_token_set=False,
                          editex=False,
                          typo=False,
                          lig_3=False,
                          ssk=False,
                          refined_soundex=False,
                          double_metaphone=False) -> dict:
    r"""
    A function which returns a dict containing the distance metrics that should be 
    used during the fuzzy string matching

    Levenshtein edit distance
        - Indel
        - Discounted Levenshtein
        - LIG3
    Block edit distances
        - Tichy
        - CormodeLZ
    Multi-set token-based distance
        - BaulieuXIII
        - Clement
        - DiceAsymmetricI
        - KuhnsIII
        - Overlap
        - PearsonII
        - WeightedJaccard
        - WarrensIV
        - Bag
        - RougeL
    Subsequence distances
        - IterativeSubString
        - RatcliffObershelp
        - SSK
    Normalized compression distance
        - NCDbz2
    FuzzyWuzzy distances
        - FuzzyWuzzyPartialString
        - FuzzyWuzzyTokenSort
        - FuzzyWuzzyTokenSet
    Ponetic distances
        - RefinedSoundex
        - DoubleMetaphone
    Edit distances
        - Editex
        - Typo


    Parameters
    ----------
    indel: bool
        Boolean indicating whether the Indel method should be used during the 
        fuzzy name matching. The indel method is equal to a regular levenshtein 
        distance with a twice as high substitution weight 
        default=False
    discounted_levenshtein: bool
        Boolean indicating whether the DiscountedLevenshtein method should be used 
        during the fuzzy name matching. Equal to the regular levenshtein distance,
        only errors later in the string are counted at a discounted rate. To for
        instance limit the importance of suffix differences
        default=False
    tichy: bool
        Boolean indicating whether the Tichy method should be used during the 
        fuzzy name matching. This algorithm provides a shortest edit distance based
        on substring and add operations.
        default=False
    cormodel_z: bool
        Boolean indicating whether the CormodeLZ method should be used during the 
        fuzzy name matching. The CormodeLZ distance between strings x and y, is the 
        minimum number of single characters or substrings of y or of the partially 
        built string which are required to produce x from left to right.
        default=False
    iterative_sub_string: bool
        Boolean indicating whether the IterativeSubString method should be used 
        during the fuzzy name matching. A method that counts the similarities 
        between two strings substrings and subtracts the differences taking into
        account the winkler similarity between the string and the substring.
        default=False
    baulieu_xiii: bool
        Boolean indicating whether the BaulieuXIII method should be used during 
        the fuzzy name matching. The Baulieu XIII distance between two strings is
        given by the following formula: (|X \ Y| + |Y \ X|) / (
        |X ∩ Y| + |X \ Y| + |Y \ X| + |X ∩ Y| ∙ (|X ∩ Y| - 4)^2)
        default=False
    clement: bool
        Boolean indicating whether the Clement method should be used during the
        fuzzy name matching. The Clement distance between two strings is given 
        by the following formula: (|X ∩ Y|/|X|)*(1-|X|/|N|) + (|(N \ X) \ Y|/|N \ X|) *
        (1-|N \ X|/|N|)
        default=False
    dice_asymmetrici: bool
        Boolean indicating whether the DiceAsymmetricI method should be used during 
        the fuzzy name matching. The Dice asymmetric similarity is given be |X ∩ Y|/|X|
        default=False
    kuhns_iii: bool
        Boolean indicating whether the KuhnsIII method should be used during the 
        fuzzy name matching
        default=False
    overlap: bool
        Boolean indicating whether the Overlap method should be used during the 
        fuzzy name matching. The overlap distance is given by: |X ∩ Y|/min(|X|,|Y|)
        default=True
    pearson_ii: bool
        Boolean indicating whether the PearsonII method should be used during the 
        fuzzy name matching. This algorithm is based on the Phi coefficient or the 
        mean square contingency
        default=False
    weighted_jaccard: bool
        Boolean indicating whether the WeightedJaccard method should be used during 
        the fuzzy name matching. This is the Jaccard distance only using a wheighing 
        for the differences of 3.
        default=True
    warrens_iv: bool
        Boolean indicating whether the WarrensIV method should be used during the 
        fuzzy name matching
        default=False
    bag: bool
        Boolean indicating whether the Bag method should be used during the fuzzy
            name matching. Is a simplification of the regular edit distance by using
            a similarity tree structure.
        default=False
    rouge_l: bool
        Boolean indicating whether the ROUGE-L method should be used during the 
        fuzzy name matching. The ROGUE-L method is a measure that counts the longest
        substring between to strings
        default=False
    ratcliff_obershelp: bool
        Boolean indicating whether the RatcliffObershelp method should be used 
        during the fuzzy name matching. This method finds the longest common substring
        and evaluates the longest common substrings to the right and the left of the 
        original longest common substring
        default=True
    ncd_bz2: bool
        Boolean indicating whether the NCDbz2 method should be used during the 
        fuzzy name matching. Applies the Burrows-Wheeler transform to the strings and 
        subsequently returns the normalised compression distance.
        default=False
    fuzzy_wuzzy_partial_string: bool
        Boolean indicating whether the FuzzyWuzzyPartialString method should be used
        during the fuzzy name matching. This methods takes the length of the longest 
        common substring and divides it over the minimum of the length of each of 
        the two strings.
        default=False
    fuzzy_wuzzy_token_sort: bool
        Boolean indicating whether the FuzzyWuzzyTokenSort method should be used 
        during the fuzzy name matching. This tokenizes the words in the string
        and sorts them, subsequently a hamming distance is calculated
        default=True
    fuzzy_wuzzy_token_set: bool
        Boolean indicating whether the FuzzyWuzzyTokenSet method should be used 
        during the fuzzy name matching. This method tokenizes the strings and 
        find the largest intersection of the two substrings and divides it over 
        the length of the shortest string
        default=False
    editex: bool
        Boolean indicating whether the Editex method should be used during the 
        fuzzy name matching
        default=True
    typo: bool
        Boolean indicating whether the Typo method should be used during the 
        fuzzy name matching. The typo distance is calculated based on the distance
        on a keyboard between edits.
        default=False
    lig_3: bool
        Boolean indicating whether the LIG3 method should be used during the fuzzy 
        name matching
        default=False
    ssk: bool
        Boolean indicating whether the SSK method should be used during the fuzzy 
        name matching. The ssk algorithm looks at the string kernel generated by all 
        the possible different subsequences present between the two strings.
        default=False
    refined_soundex: bool
        Boolean indicating whether the string should be represented by the RefinedSoundex
        phonetix translation and the Levensthein distance of the translated strings should
        be included in the fuzzy matching process
        default=False
    double_metaphone: bool
        Boolean indicating whether the string should be represented by the DoubleMetaphone
        phonetix translation and the Levensthein distance of the translated strings should
        be included in the fuzzy matching process
        default=False

    """
    distance_metrics = defaultdict(list)
    if indel:
        distance_metrics['Levenshtein'].append(Indel())
    if discounted_levenshtein:
        distance_metrics['Levenshtein'].append(
            DiscountedLevenshtein())
    if cormodel_z:
        distance_metrics['block'].append(CormodeLZ())
    if tichy:
        distance_metrics['block'].append(Tichy())
    if iterative_sub_string:
        distance_metrics['Subsequence'].append(
            IterativeSubString())
    if baulieu_xiii:
        distance_metrics['multiset'].append(BaulieuXIII())
    if clement:
        distance_metrics['multiset'].append(Clement())
    if dice_asymmetrici:
        distance_metrics['multiset'].append(DiceAsymmetricI())
    if kuhns_iii:
        distance_metrics['multiset'].append(KuhnsIII())
    if overlap:
        distance_metrics['multiset'].append(Overlap())
    if pearson_ii:
        distance_metrics['multiset'].append(PearsonII())
    if weighted_jaccard:
        distance_metrics['multiset'].append(WeightedJaccard())
    if warrens_iv:
        distance_metrics['multiset'].append(WarrensIV())
    if bag:
        distance_metrics['multiset'].append(Bag())
    if rouge_l:
        distance_metrics['multiset'].append(RougeL())
    if ratcliff_obershelp:
        distance_metrics['Subsequence'].append(
            RatcliffObershelp())
    if ncd_bz2:
        distance_metrics['compression'].append(NCDbz2())
    if fuzzy_wuzzy_partial_string:
        distance_metrics['fuzzy'].append(
            FuzzyWuzzyPartialString())
    if fuzzy_wuzzy_token_sort:
        distance_metrics['fuzzy'].append(FuzzyWuzzyTokenSort())
    if fuzzy_wuzzy_token_set:
        distance_metrics['fuzzy'].append(FuzzyWuzzyTokenSet())
    if editex:
        distance_metrics['edit'].append(Editex())
    if typo:
        distance_metrics['edit'].append(Typo())
    if lig_3:
        distance_metrics['Levenshtein'].append(LIG3())
    if ssk:
        distance_metrics['Subsequence'].append(SSK())
    if refined_soundex:
        distance_metrics['phonetic'].append(PhoneticDistance(
            transforms=RefinedSoundex(max_length=30), metric=Levenshtein(), encode_alpha=True))
    if double_metaphone:
        distance_metrics['phonetic'].append(PhoneticDistance(
            transforms=DoubleMetaphone(max_length=30), metric=Levenshtein(), encode_alpha=True))

    return distance_metrics



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._typo.

Typo edit distance functions.
"""

from itertools import chain
from math import log
from typing import Any, Dict, Tuple, cast

from numpy import float_ as np_float
from numpy import zeros as np_zeros

from ._distance import _Distance


__all__ = ['Typo']


class Typo(_Distance):
    """Typo distance.

    This is inspired by Typo-Distance :cite:`Song:2011`, and a fair bit of
    this was copied from that module. Compared to the original, this supports
    different metrics for substitution.

    .. versionadded:: 0.3.6
    """

    # fmt: off
    _keyboard = {'QWERTY': (
        (('`', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '='),
         ('', 'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '[', ']',
          '\\'),
         ('', 'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', "'"),
         ('', 'z', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', '/'),
         ('', '', '', ' ')),
        (('~', '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '_', '+'),
         ('', 'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '{', '}', '|'),
         ('', 'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', ':', '"'),
         ('', 'Z', 'X', 'C', 'V', 'B', 'N', 'M', '<', '>', '?'),
         ('', '', '', ' '))
    ), 'Dvorak': (
        (('`', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '[', ']'),
         ('', "'", ',', '.', 'p', 'y', 'f', 'g', 'c', 'r', 'l', '/', '=',
          '\\'),
         ('', 'a', 'o', 'e', 'u', 'i', 'd', 'h', 't', 'n', 's', '-'),
         ('', ';', 'q', 'j', 'k', 'x', 'b', 'm', 'w', 'v', 'z'),
         ('', '', '', ' ')),
        (('~', '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '{', '}'),
         ('', '"', '<', '>', 'P', 'Y', 'F', 'G', 'C', 'R', 'L', '?', '+', '|'),
         ('', 'A', 'O', 'E', 'U', 'I', 'D', 'H', 'T', 'N', 'S', '_'),
         ('', ':', 'Q', 'J', 'K', 'X', 'B', 'M', 'W', 'V', 'Z'),
         ('', '', '', ' '))
    ), 'AZERTY': (
        (('²', '&', 'é', '"', "'", '(', '-', 'è', '_', 'ç', 'à', ')', '='),
         ('', 'a', 'z', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '', '$'),
         ('', 'q', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', 'm', 'ù', '*'),
         ('<', 'w', 'x', 'c', 'v', 'b', 'n', ',', ';', ':', '!'),
         ('', '', '', ' ')),
        (('~', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '°', '+'),
         ('', 'A', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '', '£'),
         ('', 'Q', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', 'M', 'Ù', 'μ'),
         ('>', 'W', 'X', 'C', 'V', 'B', 'N', '?', '.', '/', '§'),
         ('', '', '', ' '))
    ), 'QWERTZ': (
        (('', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', 'ß', ''),
         ('', 'q', 'w', 'e', 'r', 't', 'z', 'u', 'i', 'o', 'p', ' ü', '+',
          '\\'),
         ('', 'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', 'ö', 'ä', '#'),
         ('<', 'y', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', '-'),
         ('', '', '', ' ')),
        (('°', '!', '"', '§', '$', '%', '&', '/', '(', ')', '=', '?', ''),
         ('', 'Q', 'W', 'E', 'R', 'T', 'Z', 'U', 'I', 'O', 'P', 'Ü', '*', ''),
         ('', 'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', 'Ö', 'Ä', "'"),
         ('>', 'Y', 'X', 'C', 'V', 'B', 'N', 'M', ';', ':', '_'),
         ('', '', '', ' '))
    ), 'QWERTZ_HUN': (
        (('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'ö', 'ü', 'ó'),
         ('', 'q', 'w', 'e', 'r', 't', 'z', 'u', 'i', 'o', 'p', ' ő', 'ú',
          '\\'),
         ('', 'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', 'é', 'á', 'ű'),
         ('í', 'y', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', '-'),
         ('', '', '', ' ')),
        (('§', "'", '"', '+', '!', '%', '/', '=', '(', ')', 'Ö', 'Ü', 'Ó'),
         ('', 'Q', 'W', 'E', 'R', 'T', 'Z', 'U', 'I', 'O', 'P', 'Ő', 'Ú', ''),
         ('', 'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', 'É', 'Á', 'Ű'),
         ('Í', 'Y', 'X', 'C', 'V', 'B', 'N', 'M', '?', ':', '_'),
         ('', '', '', ' ')),
        (('', '~', 'ˇ', '^', '˘', '°', '˛', '`', '˙', '´', '˝', '¨', '¸'),
         ('', '\\', '|', '', '', '', '', '€', '', '', '', '÷', '×', ''),
         ('', '', 'đ', 'Đ', '[', ']', '', '', 'ł', 'Ł', '$', 'ß', '¤'),
         ('<', '>', '#', '&', '@', '{', '}', '', ';', '>', '*'),
         ('', '', '', ' '))
    )}  # type: Dict[str, Tuple[Tuple[Tuple[str, ...], ...], ...]]
    # fmt: on

    def __init__(
        self,
        metric: str = 'euclidean',
        cost: Tuple[float, float, float, float] = (1.0, 1.0, 0.5, 0.5),
        layout: str = 'QWERTY',
        failsafe: bool = False,
        **kwargs: Any
    ):
        """Initialize Typo instance.

        Parameters
        ----------
        metric : str
            Supported values include: ``euclidean``, ``manhattan``,
            ``log-euclidean``, and ``log-manhattan``
        cost : tuple
            A 4-tuple representing the cost of the four possible edits:
            inserts, deletes, substitutions, and shift, respectively (by
            default: (1, 1, 0.5, 0.5)) The substitution & shift costs should be
            significantly less than the cost of an insertion & deletion unless
            a log metric is used.
        layout : str
            Name of the keyboard layout to use (Currently supported:
            ``QWERTY``, ``Dvorak``, ``AZERTY``, ``QWERTZ``, ``auto``). If
            ``auto`` is selected, the class will attempt to determine an
            appropriate keyboard based on the supplied words.
        failsafe : bool
            If True, substitution of an unknown character (one not present on
            the selected keyboard) will incur a cost equal to an insertion plus
            a deletion.
        **kwargs
            Arbitrary keyword arguments


        .. versionadded:: 0.4.0

        """
        super(Typo, self).__init__(**kwargs)
        self._metric = metric
        self._cost = cost
        self._layout = layout
        self._failsafe = failsafe

    def dist_abs(self, src: str, tar: str) -> float:
        """Return the typo distance between two strings.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            Typo distance

        Raises
        ------
        ValueError
            char not found in any keyboard layouts

        Examples
        --------
        >>> cmp = Typo()
        >>> cmp.dist_abs('cat', 'hat')
        1.5811388300841898
        >>> cmp.dist_abs('Niall', 'Neil')
        2.8251407699364424
        >>> cmp.dist_abs('Colin', 'Cuilen')
        3.414213562373095
        >>> cmp.dist_abs('ATCG', 'TAGC')
        2.5

        >>> cmp = Typo(metric='manhattan')
        >>> cmp.dist_abs('cat', 'hat')
        2.0
        >>> cmp.dist_abs('Niall', 'Neil')
        3.0
        >>> cmp.dist_abs('Colin', 'Cuilen')
        3.5
        >>> cmp.dist_abs('ATCG', 'TAGC')
        2.5

        >>> cmp = Typo(metric='log-manhattan')
        >>> cmp.dist_abs('cat', 'hat')
        0.8047189562170501
        >>> cmp.dist_abs('Niall', 'Neil')
        2.2424533248940004
        >>> cmp.dist_abs('Colin', 'Cuilen')
        2.242453324894
        >>> cmp.dist_abs('ATCG', 'TAGC')
        2.3465735902799727


        .. versionadded:: 0.3.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        ins_cost, del_cost, sub_cost, shift_cost = self._cost

        if src == tar:
            return 0.0
        if not src:
            return len(tar) * ins_cost
        if not tar:
            return len(src) * del_cost

        if self._layout == 'auto':
            for kb in ['QWERTY', 'QWERTZ', 'AZERTY']:
                keys = set(chain(*chain(*self._keyboard[kb])))
                letters = set(src) | set(tar)
                if not (letters - keys):
                    keyboard = self._keyboard[kb]
                    break
            else:
                # Fallback to QWERTY
                keyboard = self._keyboard['QWERTY']
        else:
            keyboard = self._keyboard[self._layout]

        kb_array = []
        for kb_mode in keyboard:
            kb_array.append({item for sublist in kb_mode for item in sublist})
        keys = set(chain(*chain(*keyboard)))

        def _kb_array_for_char(char: str) -> Tuple[Tuple[str, ...], ...]:
            """Return the keyboard layout that contains ch.

            Parameters
            ----------
            char : str
                The character to lookup

            Returns
            -------
            tuple
                A keyboard

            Raises
            ------
            ValueError
                char not found in any keyboard layouts

            .. versionadded:: 0.3.0

            """
            for i, kb_mode in enumerate(kb_array):
                if char in kb_mode:
                    return keyboard[i]
            raise ValueError(char + ' not found in any keyboard layouts')

        def _substitution_cost(char1: str, char2: str) -> float:
            if self._failsafe and (char1 not in keys or char2 not in keys):
                return ins_cost + del_cost
            cost = sub_cost
            cost *= metric_dict[self._metric](char1, char2) + shift_cost * (
                _kb_array_for_char(char1) != _kb_array_for_char(char2)
            )
            return cost

        def _get_char_coord(
            char: str, kb_array: Tuple[Tuple[str, ...], ...]
        ) -> Tuple[int, int]:
            """Return the row & column of char in the keyboard.

            Parameters
            ----------
            char : str
                The character to search for
            kb_array : tuple of tuples
                The array of key positions

            Returns
            -------
            tuple
                The row & column of the key

            .. versionadded:: 0.3.0

            """
            for row in kb_array:  # pragma: no branch
                if char in row:
                    break
            return kb_array.index(row), row.index(char)

        def _euclidean_keyboard_distance(char1: str, char2: str) -> float:
            row1, col1 = _get_char_coord(char1, _kb_array_for_char(char1))
            row2, col2 = _get_char_coord(char2, _kb_array_for_char(char2))
            return ((row1 - row2) ** 2 + (col1 - col2) ** 2) ** 0.5

        def _manhattan_keyboard_distance(char1: str, char2: str) -> float:
            row1, col1 = _get_char_coord(char1, _kb_array_for_char(char1))
            row2, col2 = _get_char_coord(char2, _kb_array_for_char(char2))
            return abs(row1 - row2) + abs(col1 - col2)

        def _log_euclidean_keyboard_distance(char1: str, char2: str) -> float:
            return log(1 + _euclidean_keyboard_distance(char1, char2))

        def _log_manhattan_keyboard_distance(char1: str, char2: str) -> float:
            return log(1 + _manhattan_keyboard_distance(char1, char2))

        metric_dict = {
            'euclidean': _euclidean_keyboard_distance,
            'manhattan': _manhattan_keyboard_distance,
            'log-euclidean': _log_euclidean_keyboard_distance,
            'log-manhattan': _log_manhattan_keyboard_distance,
        }

        d_mat = np_zeros((len(src) + 1, len(tar) + 1), dtype=np_float)
        for i in range(len(src) + 1):
            d_mat[i, 0] = i * del_cost
        for j in range(len(tar) + 1):
            d_mat[0, j] = j * ins_cost

        for i in range(len(src)):
            for j in range(len(tar)):
                d_mat[i + 1, j + 1] = min(
                    d_mat[i + 1, j] + ins_cost,  # ins
                    d_mat[i, j + 1] + del_cost,  # del
                    d_mat[i, j]
                    + (
                        _substitution_cost(src[i], tar[j])
                        if src[i] != tar[j]
                        else 0
                    ),  # sub/==
                )

        return cast(float, d_mat[len(src), len(tar)])

    def dist(self, src: str, tar: str) -> float:
        """Return the normalized typo distance between two strings.

        This is typo distance, normalized to [0, 1].

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            Normalized typo distance

        Examples
        --------
        >>> cmp = Typo()
        >>> round(cmp.dist('cat', 'hat'), 12)
        0.527046276695
        >>> round(cmp.dist('Niall', 'Neil'), 12)
        0.565028153987
        >>> round(cmp.dist('Colin', 'Cuilen'), 12)
        0.569035593729
        >>> cmp.dist('ATCG', 'TAGC')
        0.625


        .. versionadded:: 0.3.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        if src == tar:
            return 0.0
        ins_cost, del_cost = self._cost[:2]
        return self.dist_abs(src, tar) / (
            max(len(src) * del_cost, len(tar) * ins_cost)
        )


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.phonetic._phonetic.

The phonetic._phonetic module implements abstract class Phonetic.
"""

from itertools import groupby

__all__ = ['_Phonetic']


class _Phonetic:
    """Abstract Phonetic class.

    .. versionadded:: 0.3.6
    """

    _uc_set = set('ABCDEFGHIJKLMNOPQRSTUVWXYZ')
    _lc_set = set('abcdefghijklmnopqrstuvwxyz')
    _uc_v_set = set('AEIOU')
    _lc_v_set = set('aeiou')
    _uc_vy_set = set('AEIOUY')
    _lc_vy_set = set('aeiouy')

    def _delete_consecutive_repeats(self, word: str) -> str:
        """Delete consecutive repeated characters in a word.

        Parameters
        ----------
        word : str
            The word to transform

        Returns
        -------
        str
            Word with consecutive repeating characters collapsed to a single
            instance

        Examples
        --------
        >>> pe = _Phonetic()
        >>> pe._delete_consecutive_repeats('REDDEE')
        'REDE'
        >>> pe._delete_consecutive_repeats('AEIOU')
        'AEIOU'
        >>> pe._delete_consecutive_repeats('AAACCCTTTGGG')
        'ACTG'


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        return ''.join(char for char, _ in groupby(word))

    def encode(self, word: str) -> str:
        """Encode phonetically.

        Parameters
        ----------
        word : str
            The word to transform


        .. versionadded:: 0.3.6

        """
        return word

    def encode_alpha(self, word: str) -> str:
        """Encode phonetically using alphabetic characters.

        Parameters
        ----------
        word : str
            The word to transform

        Returns
        -------
        str
            The word transformed


        .. versionadded:: 0.3.6

        """
        return self.encode(word)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



from name_matching.name_matcher import NameMatcher
from typing import Union, Tuple
import pandas as pd
import unicodedata


def _match_names_check_data(data: Union[pd.Series, pd.DataFrame],
                            column: str,
                            group_column: str) -> pd.DataFrame:
    """
    Checks the input data of the name matching function to see whether the defined columns can 
    be found and makes a new column which will be used for the name matching
    ----------
    data: Union[pd.DataFrame, pd.Series]
        The first dataframe or series used for the name matching        
    column: str
        The column in which the name that should be matched can be found for data
    group_column_first: str
        The name of the column that should be used to generate groups within the data.

    Returns
    -------
    pd.DataFrame
        A dataframe containing data and an additional column 'name_matching_data' containing the
        names which should be matched
    """

    if isinstance(data, pd.DataFrame):
        if column == '':
            raise ValueError(
                'For one of the dataframes no column is given to perform the name matching on')
        if column not in data.columns:
            raise ValueError(
                'Could not find one of the columns in the dataframe')
        if (group_column != '') & (group_column not in data.columns):
            raise ValueError(
                'Could not find one of the group_columns in the dataframe')
        data['name_matching_data'] = data[column]
    else:
        if group_column != '':
            raise ValueError(
                'Grouping is only possible when a dataframe is used for both inputs')
        data = pd.DataFrame(data, columns=['name_matching_data'])

    return data

def _match_names_preprocess_data(column: str,
                                data_first: pd.DataFrame,
                                data_second: pd.DataFrame,
                                case_sensitive: bool,
                                punctuation_sensitive: bool,
                                special_character_sensitive: bool) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Preprocess the data by making the names lower case, removing punctuations and special characters. 
    And convert the indexes of the second dataframe to a column.
    ----------
    data: Union[pd.DataFrame, pd.Series]
        The first dataframe or series used for the name matching        
    column: str
        The column in which the name that should be matched can be found for data
    group_column_first: str
        The name of the column that should be used to generate groups within the data.
    case_sensitive: bool
        Boolean value indicating whether the names should be converted to lower case names
        before the name matching starts. If False all the characters are converted to lowercase
    punctuation_sensitive: bool
        Boolean value indicating whether punctuations should be removed from the original names
        before the name matching starts. If False the punctuations are removed
    special_character_sensitive: bool
        Boolean value indicating whether special characters should be converted to unicode
        before the name matching starts. If False the special characters are replaced

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        A dataframe containing data and an additional column 'name_matching_data' containing the
        names which should be matched
    """

    if not case_sensitive:
        data_first[column] = data_first[column].str.lower().str.strip()
        data_second[column] = data_second[column].str.lower().str.strip()
    if not punctuation_sensitive:
        data_first[column] = data_first[column].str.replace('[^\w\s]', '', regex=True)
        data_second[column] = data_second[column].str.replace(
            '[^\w\s]', '', regex=True)
    if not special_character_sensitive:
        data_first[column] = data_first[column].apply(lambda string: unicodedata.normalize(
            'NFKD', string).encode('ASCII', 'ignore').decode())
        data_second[column] = data_second[column].apply(lambda string: unicodedata.normalize(
            'NFKD', string).encode('ASCII', 'ignore').decode())

    data_second = data_second.rename_axis('index').reset_index(drop=False)
    
    return data_first, data_second

def _match_names_combine_data(data_first: pd.DataFrame, 
                              data_second: pd.DataFrame, 
                              left_cols: list, 
                              right_cols: list) -> pd.DataFrame:
    """
    Perform a merge to match data based on whether the names are equal
    ----------
    data_first: pd.DataFrame
        The first dataframe or series used for the name matching        
    data_second: pd.DataFrame
        The second dataframe or series used for the name matching        
    left_cols: list
        A list of columns on which the first dataframe should be merged  
    right_cols: list
        A list of columns on which the first dataframe should be merged

    Returns
    -------
    pd.DataFrame
        A dataframe containing the original name, matched name, match score and match index. The index of the
        dataframe is equal to the original index of data_first, the match index is the index in data_second
        for the matched name.
    """
    matches = pd.merge(data_first, data_second, how='left',
                        left_on=left_cols, right_on=right_cols, suffixes=['', '_matched'])
    matches['score'] = 100
    matches = matches.dropna(subset=['index'])
    matches = matches.rename(columns={'index':'match_index'})
    matches = matches[['match_index', 'score']] 

    return matches

def _match_names_match_single(matcher: NameMatcher,
                              data_first: pd.DataFrame,
                              data_second: pd.DataFrame,
                              name_column: str) -> pd.DataFrame:
    """
    Perform the name matching. First by doing a perfect string match with a merge statement, followed
    by the fuzzy matching approach as done in NameMatcher. 
    ----------
    matcher: NameMatcher
        The NameMatcher to be used for the name matching part        
    data_first: pd.DataFrame
        The first dataframe or series used for the name matching        
    data_second: pd.DataFrame
        The second dataframe or series used for the name matching        
    name_column: str
        The column in which the name that should be matched can be found for both dataframes

    Returns
    -------
    pd.DataFrame
        A dataframe containing the original name, matched name, match score and match index. The index of the
        dataframe is equal to the original index of data_first, the match index is the index in data_second
        for the matched name.
    """

    matches = _match_names_combine_data(data_first, data_second,
                            [name_column], [name_column])
    unmatched = data_first[~data_first.index.isin(matches.index)].copy()
    if len(unmatched) > 0:
        matcher.load_and_process_master_data(name_column, data_second, transform=True)
        matches = pd.concat([matches,(matcher.match_names(
            to_be_matched=unmatched, column_matching=name_column))])
        return matches
    else:
        print('All data matched with basic string matching')
        return matches

def _match_names_match_group(matcher: NameMatcher,
                            data_first: pd.DataFrame,
                            data_second: pd.DataFrame,
                            name_column: str,
                            group_column_first: str,
                            group_column_second: str) -> pd.DataFrame:
    """
    Perform the name matching based on the subgroups as indicated by the group_column strings. First by doing
    a perfect string match with a merge statement, followed by the fuzzy matching approach as done in NameMatcher. 
    ----------
    matcher: NameMatcher
        The NameMatcher to be used for the name matching part        
    data_first: pd.DataFrame
        The first dataframe or series used for the name matching        
    data_second: pd.DataFrame
        The second dataframe or series used for the name matching        
    name_column: str
        The column in which the name that should be matched can be found for both dataframes
    group_column_first: str
        The name of the column that should be used to generate groups within the data.
    group_column_second: str
        The name of the column that should be used to generate groups within the data.

    Returns
    -------
    pd.DataFrame
        A dataframe containing the original name, matched name, match score and match index. The index of the
        dataframe is equal to the original index of data_first, the match index is the index in data_second
        for the matched name.
    """

    matches = _match_names_combine_data(data_first, data_second, [
                            name_column, group_column_first], [name_column, group_column_second])
    unmatched = data_first[~data_first.index.isin(matches.index)]
    if len(unmatched) > 0:
        matcher.load_and_process_master_data(name_column, data_second, transform=False)
        for group in data_first[group_column_first].unique():
            data_second_group = data_second[data_second[group_column_second] == group].copy()
            matcher.load_and_process_master_data(name_column, 
                data_second_group, start_processing=False)
            matcher.transform_data()
            matches = pd.concat([matches, matcher.match_names(
                to_be_matched=unmatched[unmatched[group_column_first] == group].copy(), column_matching=name_column)])
    else:
        print('All data matched with basic string matching')
        return matches

    return matches


def match_names(data_first: Union[pd.DataFrame, pd.Series],
                data_second: Union[pd.DataFrame, pd.Series],
                column_first='',
                column_second='',
                group_column_first='',
                group_column_second='',
                case_sensitive=False,
                punctuation_sensitive=False,
                special_character_sensitive=False,
                threshold=95,
                **kwargs) -> pd.DataFrame:
    """Function which performs name matching. First a simple merge on the data is performed
    to get the instances in which the name matches perfectly. Subsequently the matches are
    matched using the name matching algorithm as defined in name_matcher.

    Parameters
    ----------
    data_first: Union[pd.DataFrame, pd.Series]
        The first dataframe or series used for the name matching
    data_second: Union[pd.DataFrame, pd.Series]
        The second dataframe or series used for the name matching, for matching the data to 
        itself data_second should be equal to data first
    column_first: str
        If data_first is a dataframe column_first should be the column in which the name 
        that should be matched can be found for data_first
        default=''
    column_second: str
        If data_second is a dataframe column_second should be the column in which the name 
        that should be matched can be found for data_second
        default=''
    group_column_first: str
        The name of the column that should be used to generate groups within the data_first 
        dataframe. The matchig is then only performed for instances in which the groups are
        identical
        default=''
    group_column_second: str
        The name of the column that should be used to generate groups within the data_second 
        dataframe. The matchig is then only performed for instances in which the groups are
        identical
        default=''
    case_sensitive: bool
        Boolean value indicating whether the names should be converted to lower case names
        before the name matching starts. If False all the characters are converted to lowercase
        default=False
    punctuation_sensitive: bool
        Boolean value indicating whether punctuations should be removed from the original names
        before the name matching starts. If False the punctuations are removed
        default=False
    special_character_sensitive: bool
        Boolean value indicating whether special characters should be converted to unicode
        before the name matching starts. If False the special characters are replaced
        default=False
    threshold: int
        the minimal score a match should have to be part of the output
        default=95
    **kwargs
        Additional inputs for the name_matcher

    Returns
    -------
    pd.DataFrame
        A dataframe containing the matched rows were the match score is above the threshold. The
        dataframe consists of 4 columns; original_name: the original name from data_first after 
        preprocessing, match_name_0: the name it is matched to from data_second after preprocessing,
        score_0: the score of the match, match_index_0: the index of the match in data_second. The 
        match_index_0 can be used to join the data from both dataframes. 
    """
    if 'number_of_matches' in kwargs:
        raise ValueError(
            'The number of matches can only be changed by using a custom matching approach')

    data_first = _match_names_check_data(data_first, column_first, group_column_first)
    data_second = _match_names_check_data(data_second, column_second, group_column_second)

    name_column = 'name_matching_data'

    if ((group_column_first == '') & (group_column_second != '')) | ((group_column_second == '') & (group_column_first != '')):
        raise ValueError(
            'For the grouping to work both the grouping column in the first as well as the second dataframe have to be indicated')

    if (threshold > 100) | (threshold < 0):
        raise ValueError('Please pick a threshold between 0 and 100')

    data_first, data_second = _match_names_preprocess_data(name_column, data_first, 
            data_second, case_sensitive, punctuation_sensitive, special_character_sensitive)

    matcher = NameMatcher(**kwargs)

    if group_column_first == '':
        matches = _match_names_match_single(matcher, data_first, data_second, name_column)
    else:
        matches = _match_names_match_group(matcher, data_first, data_second,
                              name_column, group_column_first, group_column_second)

    return matches[matches['score'] > threshold]



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._lcprefix.

Longest common prefix
"""

from os.path import commonprefix
from typing import List, cast

from ._distance import _Distance

__all__ = ['LCPrefix']


class LCPrefix(_Distance):
    """Longest common prefix.

    .. versionadded:: 0.4.0
    """

    def lcprefix(self, strings: List[str]) -> str:
        """Return the longest common prefix of a list of strings.

        Longest common prefix (LCPrefix).

        Parameters
        ----------
        strings : list of strings
            Strings for comparison

        Returns
        -------
        str
            The longest common prefix

        Examples
        --------
        >>> pfx = LCPrefix()
        >>> pfx.lcprefix(['cat', 'hat'])
        ''
        >>> pfx.lcprefix(['Niall', 'Neil'])
        'N'
        >>> pfx.lcprefix(['aluminum', 'Catalan'])
        ''
        >>> pfx.lcprefix(['ATCG', 'TAGC'])
        ''


        .. versionadded:: 0.4.0

        """
        return cast(str, commonprefix(strings))

    def dist_abs(self, src: str, tar: str, *args: str) -> int:
        """Return the length of the longest common prefix of the strings.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison
        *args : strs
            Additional strings for comparison

        Raises
        ------
        ValueError
            All arguments must be of type str

        Returns
        -------
        int
            The length of the longest common prefix

        Examples
        --------
        >>> pfx = LCPrefix()
        >>> pfx.dist_abs('cat', 'hat')
        0
        >>> pfx.dist_abs('Niall', 'Neil')
        1
        >>> pfx.dist_abs('aluminum', 'Catalan')
        0
        >>> pfx.dist_abs('ATCG', 'TAGC')
        0


        .. versionadded:: 0.4.0

        """
        strings = [src, tar]
        for arg in args:
            if isinstance(arg, str):
                strings.append(arg)
            else:
                raise TypeError('All arguments must be of type str')

        return len(self.lcprefix(strings))

    def sim(self, src: str, tar: str, *args: str) -> float:
        r"""Return the longest common prefix similarity of two or more strings.

        Longest common prefix similarity (:math:`sim_{LCPrefix}`).

        This employs the LCPrefix function to derive a similarity metric:
        :math:`sim_{LCPrefix}(s,t) = \frac{|LCPrefix(s,t)|}{max(|s|, |t|)}`

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison
        *args : strs
            Additional strings for comparison

        Returns
        -------
        float
            LCPrefix similarity

        Examples
        --------
        >>> pfx = LCPrefix()
        >>> pfx.sim('cat', 'hat')
        0.0
        >>> pfx.sim('Niall', 'Neil')
        0.2
        >>> pfx.sim('aluminum', 'Catalan')
        0.0
        >>> pfx.sim('ATCG', 'TAGC')
        0.0


        .. versionadded:: 0.4.0

        """
        if src == tar:
            return 1.0
        elif not src or not tar:
            return 0.0
        dist = self.dist_abs(src, tar, *args)
        maxlen = max(len(src), len(tar), *[len(arg) for arg in args])
        return dist / maxlen


if __name__ == '__main__':
    import doctest

    doctest.testmod()



import unicodedata
import functools
import operator
import re
import numpy as np
import pandas as pd
from tqdm import tqdm
from typing import Union, Tuple
from itertools import compress
from sklearn.feature_extraction.text import TfidfVectorizer
from name_matching.distance_metrics import make_distance_metrics
from cleanco.termdata import terms_by_type, terms_by_country
from name_matching.sparse_cosine import sparse_cosine_top_n


class NameMatcher:

    """
    A class for the name matching of data based on the strings in a single column. The NameMatcher
    first applies a cosine similarity on the ngrams of the strings to get an approximate match followed 
    by a fuzzy matching based on a number of different algorithms.

    Parameters
    ----------
    ngrams : tuple of integers
        The length of the ngrams which should be used for the generation of ngrams for the cosine
        similarity comparison of the possible matches
        default=(2, 3)
    top_n : integer
        The number of possible matches that should be included in the group which will be analysed
        with the fuzzy matching algorithms
        default=50
    low_memory : bool
        Bool indicating if the a low memory approach should be taken in the sparse cosine similarity
        step.
        default=False
    number_of_rows : integer
        Determines how many rows should be calculated at once with the sparse cosine similarity step.
        If the low_memory bool is True this number is unused.
        default=5000
    number_of_matches : int
        The number of matches which should be returned by the matching algorithm. If a number higher
        than 1 is given, a number of alternative matches are also returned. If the number is equal
        to the number of algorithms used, the best match for each algorithm is returned. If the
        number is equal to the number of algorithm groups which are included the best match for each
        group is returned.
        default=1
    legal_suffixes : bool
        Boolean indicating whether the most common company legal terms should be excluded when calculating 
        the final score. The terms are still included in determining the best match.
        default=False
    common_words : bool
        Boolean indicating whether the most common words from the matching data should be excluded 
        when calculating the final score. The terms are still included in determining the best match.
        default=False
    cut_off_no_scoring_words: float
        the cut off percentage of the occurrence of the most occurring word for which words are still included 
        in the no_soring_words set
        default=0.01
    lowercase : bool
        A boolean indicating whether during the preprocessing all characters should be converted to
        lowercase, to generate case insensitive matching
        default=True
    punctuations : bool
        A boolean indicating whether during the preprocessing all punctuations should be ignored
        default=True
    remove_ascii : bool
        A boolean indicating whether during the preprocessing all characters should be converted to
        ascii characters
        default=True : bool
    preprocess_split
        Indicating whether during the preprocessing an additional step should be taken in which only 
        the most common words out of a name are isolated and used in the matching process. The removing 
        of the common words is only done for the n-grams cosine matching part.
        default=False
    verbose : bool
        A boolean indicating whether progress printing should be done
        default=True
    distance_metrics: list
        A list of The distance metrics to be used during the fuzzy matching. For a list of possible distance
        metrics see the distance_metrics.py file. By default the following metrics are used: overlap, weighted_jaccard, 
                ratcliff_obershelp, fuzzy_wuzzy_token_sort and editex.

    """

    def __init__(self,
                 ngrams: tuple = (2, 3),
                 top_n: int = 50,
                 low_memory: bool = False,
                 number_of_rows: int = 5000,
                 number_of_matches: int = 1,
                 lowercase: bool = True,
                 punctuations: bool = True,
                 remove_ascii: bool = True,
                 legal_suffixes: bool = False,
                 common_words: bool = False,
                 cut_off_no_scoring_words: float = 0.01,
                 preprocess_split: bool = False,
                 verbose: bool = True,
                 distance_metrics: Union[list, tuple] = ['overlap', 'weighted_jaccard', 'ratcliff_obershelp',
                                                         'fuzzy_wuzzy_token_sort', 'editex']):

        self._possible_matches = None
        self._preprocessed = False
        self._word_set = set()
        self._df_matching_data = pd.DataFrame()

        self._number_of_rows = number_of_rows
        self._low_memory = low_memory

        self._column = ''
        self._column_matching = ''

        self._verbose = verbose
        self._number_of_matches = number_of_matches
        self._top_n = top_n

        self._preprocess_lowercase = lowercase
        self._preprocess_punctuations = punctuations
        self._preprocess_ascii = remove_ascii
        self._postprocess_company_legal_id = legal_suffixes
        self._postprocess_common_words = common_words

        self._preprocess_common_words = pd.Series(dtype=float)
        self._preprocess_split = preprocess_split
        self._cut_off = cut_off_no_scoring_words

        if self._postprocess_company_legal_id:
            self._word_set = self._make_no_scoring_words(
                'legal', self._word_set, self._cut_off)

        self.set_distance_metrics(distance_metrics)

        self._vec = TfidfVectorizer(
            lowercase=False, analyzer="char", ngram_range=(ngrams))
        self._n_grams_matching = None

    def set_distance_metrics(self, metrics: list) -> None:
        """
        A method to set which of the distance metrics should be employed during the
        fuzzy matching. For very short explanations of most of the name matching 
        algorithms please see the make_distance_metrics function in distance_matrics.py

        Parameters
        ----------
        metrics: list 
            The list with the distance metrics to be used during the name matching. The
            distance metrics can be chosen from the list below:
                indel
                discounted_levenshtein
                tichy
                cormodeL_z
                iterative_sub_string
                baulieu_xiii
                clement
                dice_asymmetricI
                kuhns_iii
                overlap
                pearson_ii
                weighted_jaccard
                warrens_iv
                bag
                rouge_l
                ratcliff_obershelp
                ncd_bz2
                fuzzy_wuzzy_partial_string
                fuzzy_wuzzy_token_sort
                fuzzy_wuzzy_token_set
                editex
                typo
                lig_3
                ssk
                refined_soundex
                double_metaphone
        """

        input_metrics = {str(metric).lower(): True for metric in metrics}
        try:
            self._distance_metrics = make_distance_metrics(**input_metrics)
        except TypeError:
            raise TypeError('Not all of the supplied distance metrics are available. Please check the' +
                            'list of options in the make_distance_metrics function and adjust your list accordingly')
        self._num_distance_metrics = sum(
            [len(x) for x in self._distance_metrics.values()])

    def _select_top_words(self, word: str, word_counts: pd.Series, occurrence_count: int) -> str:
        """Remove the top words from the string word based on an occurrence_count threshold

        Parameters
        ----------
        word: str 
            the string from which the words should be removed
        word_counts: pd.Series
            the words which should be removed with their counts as result from a value_counts
        occurrence_count: int 
            the multiplication factor of the minimum occurrences below which to select 

        Returns
        -------
        str
           The string word with the words with a too high word_counts removed
        """
        compressed_list = list(compress(
            word, (word_counts[word] < occurrence_count*word_counts[word].min()).values))

        return ' '.join(compressed_list)

    def _preprocess_reduce(self,
                           to_be_matched: pd.DataFrame,
                           occurrence_count: int = 3) -> pd.DataFrame:
        """Preprocesses and copies the data to obtain the data with reduced strings. The strings have all words
        removed which appear more than 3x as often as the least common word in the string and returns an adjusted 
        copy of the input 

        Parameters
        ----------
        to_be_matched: pd.DataFrame 
            A dataframe from which the most common words should be removed
        occurrence_count: int
            The number of occurrence a word can occur more then the least common word in the string for which it will
            still be included in the process 
            default=3

        Returns
        -------
        pd.DataFrame
            A dataframe that will contain the reduced strings
        """
        individual_words = to_be_matched[self._column_matching].str.split(
            expand=True).stack()
        word_counts = individual_words.value_counts()
        preprocess_common_words_inst = pd.concat([self._preprocess_common_words,
            word_counts]).fillna(0)
        to_be_matched_new = to_be_matched.copy()
        name = to_be_matched[self._column_matching].str.split()
        to_be_matched_new[self._column_matching] = name.apply(
            lambda word: self._select_top_words(word, preprocess_common_words_inst, occurrence_count))

        return to_be_matched_new

    def load_and_process_master_data(self,
                                     column: str,
                                     df_matching_data: pd.DataFrame,
                                     start_processing: bool = True,
                                     transform: bool = True) -> None:
        """Load the matching data into the NameMatcher and start the preprocessing.

        Parameters
        ----------
        column : string
            The column name of the dataframe which should be used for the matching
        df_matching_data: pd.DataFrame
            The dataframe which is used to match the data to.
        start_processing : bool 
            A boolean indicating whether to start the preprocessing step after loading the matching data 
            default: True
        transform : bool 
            A boolean indicating whether or not the data should be transformed after the vectoriser is initialised 
            default: True
        """
        self._column = column
        self._df_matching_data = df_matching_data
        if start_processing:
            self._process_matching_data(transform)

    def _process_matching_data(self,
                               transform: bool = True) -> None:
        """Function to process the matching data. First the matching data is preprocessed and assigned to 
        a variable within the NameMatcher. Next the data is used to initialise the TfidfVectorizer. 

        Parameters
        ----------
        transform : bool 
            A boolean indicating whether or not the data should be transformed after the vectoriser is initialised 
            default: True
        """
        self._df_matching_data = self.preprocess(
            self._df_matching_data, self._column)
        if self._postprocess_common_words:
            self._word_set = self._make_no_scoring_words(
                'common', self._word_set, self._cut_off)
        self._vectorise_data(transform)
        self._preprocessed = True

    def match_names(self,
                    to_be_matched: Union[pd.Series, pd.DataFrame],
                    column_matching: str) -> Union[pd.Series, pd.DataFrame]:
        """Performs the name matching operation on the to_be_matched data. First it does the preprocessing of the 
        data to be matched as well as the matching data if this has not been performed. Subsequently based on 
        ngrams a cosine similarity is computed between the matching data and the data to be matched, to the top n
        matches fuzzy matching algorithms are performed to determine the best match and the quality of the match

        Parameters
        ----------
        to_be_matched: Union[pd.Series, pd.DataFrame]
            The data which should be matched
        column_matching: str
            string indicating the column which will be matched

        Returns
        -------
        Union[pd.Series, pd.DataFrame]
            A series or dataframe depending on the input containing the match index from the matching_data dataframe. 
            the name in the to_be_matched data, the name to which the datapoint was matched and a score between 0 
            (no match) and 100(perfect match) to indicate the quality of the matches
        """
        if self._column == '':
            raise ValueError(
                'Please first load the master data via the method: load_and_process_master_data')
        if self._verbose:
            tqdm.pandas()
            tqdm.write('preprocessing...\n')
        self._column_matching = column_matching

        if isinstance(to_be_matched, pd.Series):
            to_be_matched = pd.DataFrame(
                [to_be_matched.values], columns=to_be_matched.index.to_list())
        if not self._preprocessed:
            self._process_matching_data()
        to_be_matched = self.preprocess(to_be_matched, self._column_matching)

        if self._verbose:
            tqdm.write('preprocessing complete \n searching for matches...\n')

        self._possible_matches = self._search_for_possible_matches(
            to_be_matched)

        if self._preprocess_split:
            self._possible_matches = np.hstack((self._search_for_possible_matches(
                self._preprocess_reduce(to_be_matched)), self._possible_matches))

        if self._verbose:
            tqdm.write('possible matches found   \n fuzzy matching...\n')
            data_matches = to_be_matched.progress_apply(lambda x: self.fuzzy_matches(
                self._possible_matches[to_be_matched.index.get_loc(x.name), :], x), axis=1)
        else:
            data_matches = to_be_matched.apply(lambda x: self.fuzzy_matches(
                self._possible_matches[to_be_matched.index.get_loc(x.name), :], x), axis=1)

        if self._number_of_matches == 1:
            data_matches = data_matches.rename(columns={'match_name_0': 'match_name',
                                                        'score_0': 'score', 'match_index_0': 'match_index'})

        if self._verbose:
            tqdm.write('done')

        return data_matches

    def fuzzy_matches(self,
                      possible_matches: np.array,
                      to_be_matched: pd.Series) -> pd.Series:
        """ A method which performs the fuzzy matching between the data in the to_be_matched series as well
        as the indicated indexes of the matching_data points which are possible matching candidates.

        Parameters
        ----------
        possible_matches : np.array
            An array containing the indexes of the matching data with potential matches
        to_be_matched : pd.Series
            The data which should be matched

        Returns
        -------
        pd.Series
            A series containing the match index from the matching_data dataframe. the name in the to_be_matched data,
            the name to which the datapoint was matched and a score between 0 (no match) and 100(perfect match) to 
            indicate the quality of the matches
        """
        if len(possible_matches.shape) > 1:
            possible_matches = possible_matches[0]

        indexes = np.array([[f'match_name_{num}', f'score_{num}', f'match_index_{num}']
                            for num in range(self._number_of_matches)]).flatten()
        match = pd.Series(index=np.append('original_name', indexes), dtype=float)
        match['original_name'] = to_be_matched[self._column_matching]
        list_possible_matches = self._df_matching_data.iloc[
            possible_matches.flatten(), :][self._column].values

        match_score = self._score_matches(
            to_be_matched[self._column_matching], list_possible_matches)
        ind = self._rate_matches(match_score)

        for num, col_num in enumerate(ind):
            match[f'match_name_{num}'] = list_possible_matches[col_num]
            match[f'match_index_{num}'] = possible_matches[col_num]

        match = self._adjust_scores(match_score[ind, :], match)

        if self._postprocess_common_words or self._postprocess_company_legal_id:
            match = self.postprocess(match)

        return match

    def _score_matches(self,
                       to_be_matched_instance: str,
                       possible_matches: list) -> np.array:
        """A method to score a name to_be_matched_instance to a list of possible matches. The scoring is done
        based on all the metrics which are enabled.

        Parameters
        ----------
        to_be_matched_instance : str
            The name which should match one of the possible matches
        possible_matches : list
            list of the names of the possible matches

        Returns
        -------
        np.array
            The score of each of the matches with respect to the different metrics which are assessed.
        """
        match_score = np.zeros(
            (len(possible_matches), self._num_distance_metrics))
        idx = 0
        for method_list in self._distance_metrics.values():
            for method in method_list:
                match_score[:, idx] = np.array(
                    [method.sim(to_be_matched_instance, s) for s in possible_matches])
                idx = idx + 1

        return match_score

    def _rate_matches(self,
                      match_score: np.array) -> np.array:
        """Converts the match scores from the score_matches method to a list of indexes of the best scoring 
        matches limited to the _number_of_matches.

        Parameters
        ----------
        match_score : np.array
            An array containing the scores of each of the possible alternatives for each
            of the different methods used

        Returns
        -------
        np.array
            The indexes of the best rated matches
        """
        if self._number_of_matches == 1:
            ind = [np.argmax(np.mean(match_score, axis=1))]
        elif self._number_of_matches == len(self._distance_metrics):
            ind = np.zeros(len(self._distance_metrics))
            idx = 0
            for num, method_list in enumerate(self._distance_metrics.values()):
                method_grouped_results = np.reshape(
                    match_score[:, idx: idx + len(method_list)], (-1, len(method_list)))
                ind[num] = np.argmax(np.mean(method_grouped_results, axis=1))
                idx = idx + len(method_list)
        elif self._number_of_matches == self._num_distance_metrics:
            ind = np.argmax(match_score, axis=1)
        else:
            ind = np.argsort(np.mean(match_score, axis=1)
                             )[-self._number_of_matches:][::-1]

        return np.array(ind, dtype=int)

    def _get_alternative_names(self, match: pd.Series) -> list:
        """Gets all the possible match names from the match.

        Parameters
        ----------
        match : pd.Series
            The series with the possible matches and original scores

        Returns
        -------
        list
            A list with the alternative names for this match
        """
        alt_names = []

        for num in range(self._number_of_matches):
            alt_names.append(str(match[f'match_name_{num}']))

        return alt_names

    def _process_words(self, org_name: str, alt_names: list) -> Tuple[str, list]:
        """Removes the words from the word list from the org_name and all the names in alt_names .

        Parameters
        ----------
        org_name : str
            The original name for the matching data
        alt_names : list
            A list of names from which the words should be removed

        Returns
        -------
        Tuple[str, list]
            The processed version of the org_name and the alt_names, with the words removed
        """
        len_atl_names = len(alt_names)
        for word in self._word_set:
            org_name = ' '.join(
                re.sub(fr'\b{re.escape(word)}\b', '', org_name).split())
            for num in range(len_atl_names):
                alt_names[num] = ' '.join(
                    re.sub(fr'\b{re.escape(word)}\b', '', alt_names[num]).split())

        return org_name, alt_names

    def _adjust_scores(self, match_score: np.array, match: pd.Series) -> pd.Series:
        """Adjust the scores to be between 0 and 100

        Parameters
        ----------
        match_score : np.array
            An array with the scores for each of the options
        match : pd.Series
            The series with the possible matches and original scores

        Returns
        -------
        pd.Series
            The series with the possible matches and adjusted scores
        """
        for num in range(self._number_of_matches):
            match[f'score_{num}'] = 100*np.mean(match_score[num, :])

        return match

    def postprocess(self,
                    match: pd.Series) -> pd.Series:
        """Postprocesses the scores to exclude certain specific company words or the most 
        common words. In this method only the scores are adjusted, the matches still stand.

        Parameters
        ----------
        match : pd.Series
            The series with the possible matches and original scores

        Returns
        -------
        pd.Series
            A new version of the input series with updated scores
        """
        alt_names = self._get_alternative_names(match)
        org_name = str(match['original_name'])

        org_name, alt_names = self._process_words(org_name, alt_names)

        match_score = self._score_matches(org_name, alt_names)

        match = self._adjust_scores(match_score, match)

        return match

    def _vectorise_data(self,
                        transform: bool = True):
        """Initialises the TfidfVectorizer, which generates ngrams and weights them based on the occurrance.
        Subsequently the matching data will be used to fit the vectoriser and the matching data might also be send
        to the transform_data function depending on the transform boolean.

        Parameters
        ----------
        transform : bool 
            A boolean indicating whether or not the data should be transformed after the vectoriser is initialised 
            default: True
        """
        self._vec.fit(self._df_matching_data[self._column].values.flatten())
        if transform:
            self.transform_data()

    def transform_data(self):
        """A method which transforms the matching data based on the ngrams transformer. After the 
        transformation (the generation of the ngrams), the data is normalised by dividing each row
        by the sum of the row. Subsequently the data is changed to a coo sparse matrix format with
        the column indices in ascending order.
        """
        ngrams = self._vec.transform(
            self._df_matching_data[self._column].astype(str))
        for i, j in zip(ngrams.indptr[:-1], ngrams.indptr[1:]):
            ngrams.data[i:j] = ngrams.data[i:j]/np.sum(ngrams.data[i:j])
        self._n_grams_matching = ngrams.tocsc()
        if self._low_memory:
            self._n_grams_matching = self._n_grams_matching.tocoo()

    def _search_for_possible_matches(self,
                                     to_be_matched: pd.DataFrame) -> np.array:
        """Generates ngrams from the data which should be matched, calculate the cosine simularity
        between these data and the matching data. Hereafter a top n of the matches is selected and
        returned.

        Parameters
        ----------
        to_be_matched : pd.Series
            A series containing a single instance of the data to be matched

        Returns
        -------
        np.array
            An array of top n values which are most closely matched to the to be matched data based
            on the ngrams
        """
        if self._n_grams_matching is None:
            raise RuntimeError(
                """First the data needs to be transformed to be able to use the sparse cosine simularity. To""" +
                """transform the data, run transform_data or run load_and_process_master_data with transform=True""")

        if self._low_memory:
            results = np.zeros((len(to_be_matched), self._top_n))
            input_data = to_be_matched[self._column_matching]
            for idx, row_name in enumerate(tqdm(input_data, disable=not self._verbose)):
                match_ngrams = self._vec.transform([row_name])
                results[idx, :] = sparse_cosine_top_n(
                    matrix_a=self._n_grams_matching, matrix_b=match_ngrams, top_n=self._top_n, low_memory=self._low_memory, number_of_rows=self._number_of_rows, verbose=self._verbose)
        else:
            match_ngrams = self._vec.transform(
                to_be_matched[self._column_matching].tolist()).tocsc()
            results = sparse_cosine_top_n(
                matrix_a=self._n_grams_matching, matrix_b=match_ngrams, top_n=self._top_n, low_memory=self._low_memory, number_of_rows=self._number_of_rows, verbose=self._verbose)

        return results

    def preprocess(self,
                   df: pd.DataFrame,
                   column_name: str) -> pd.DataFrame:
        """Preprocess a dataframe before applying a name matching algorithm. The preprocessing consists of 
        removing special characters, spaces, converting all characters to lower case and removing the
        words given in the word lists

        Parameters
        ----------
        df : DataFrame
            The dataframe or series on which the preprocessing needs to be performed        
        column_name : str
            The name of the column that is used for the preprocessing

        Returns
        -------
        pd.DataFrame
            The preprocessed dataframe or series depending on the input
        """
        df.loc[:, column_name] = df[column_name].astype(str)
        if self._preprocess_lowercase:
            df.loc[:, column_name] = df[column_name].str.lower()
        if self._preprocess_punctuations:
            df.loc[:, column_name] = df[column_name].str.replace(
                '[^\w\s]', '', regex=True)
            df.loc[:, column_name] = df[column_name].str.replace(
                '  ', ' ').str.strip()
        if self._preprocess_ascii:
            df.loc[:, column_name] = df[column_name].apply(lambda string: unicodedata.normalize(
                'NFKD', str(string)).encode('ASCII', 'ignore').decode())

        return df

    def _preprocess_word_list(self, terms: dict) -> list:
        """Preprocess legal words to remove punctuations and trailing leading space

        Parameters
        -------
        terms: dict
            a dictionary of legal words

        Returns
        -------
        list
            A list of preprocessed legal words  
        """
        if self._preprocess_punctuations:
            return [re.sub(r'[^\w\s]', '', s).strip() for s in functools.reduce(
                operator.iconcat, terms.values(), [])]
        else:
            return [s.strip() for s in functools.reduce(
                    operator.iconcat, terms.values(), [])]

    def _process_legal_words(self, word_set: set) -> set:
        """Preprocess legal words and add them to the word_set

        Parameters
        -------
        word_set: str
            the current word list which should be extended with additional words

        Returns
        -------
        Set
            The original word_set with the legal words added
        """
        terms_type = self._preprocess_word_list(terms_by_type)
        terms_country = self._preprocess_word_list(terms_by_country)
        word_set = word_set.union(set(terms_country + terms_type))

        return word_set

    def _process_common_words(self, word_set: set, cut_off: float) -> set:
        """A method to select the most common words from the matching_data.

        Parameters
        -------
        word_set: str
            the current word list which should be extended with additional words  
        cut_off: float
            the cut_off percentage of the occurrence of the most occurring word for which words are still included 
            in the no_soring_words set

        Returns
        -------
        Set
            The current word set with the most common words from the matching_data added
        """
        word_counts = self._df_matching_data[self._column].str.split(
            expand=True).stack().value_counts()
        word_set = word_set.union(
            set(word_counts[word_counts > np.max(word_counts)*cut_off].index))

        return word_set

    def _make_no_scoring_words(self,
                               indicator: str,
                               word_set: set,
                               cut_off: float) -> set:
        """A method to make a set of words which are not taken into account when scoring matches.

        Parameters
        -------
        indicator: str
            indicator for which types of words should be excluded can be legal for
            legal suffixes or common for the most common words
        word_set: str
            the current word list which should be extended with additional words  
        cut_off: float
            the cut_off percentage of the occurrence of the most occurring word for which words are still included 
            in the no_soring_words set

        Returns
        -------
        Set
            The set of no scoring words
        """
        if indicator == 'legal':
            word_set = self._process_legal_words(word_set)
        if indicator == 'common':
            word_set = self._process_common_words(word_set, cut_off)

        return word_set



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._bag.

Bag similarity & distance
"""
from typing import Any, Optional

from ._token_distance import _TokenDistance
from ._tokenizer import _Tokenizer
from ._character import CharacterTokenizer

__all__ = ['Bag']


class Bag(_TokenDistance):
    """Bag distance.

    Bag distance is proposed in :cite:`Bartolini:2002`. It is defined as

        .. math::

            dist_{bag}(src, tar) =
            max(|multiset(src)-multiset(tar)|, |multiset(tar)-multiset(src)|)

    .. versionadded:: 0.3.6
    """

    def __init__(
        self,
        tokenizer: Optional[_Tokenizer] = None,
        intersection_type: str = 'crisp',
        **kwargs: Any
    ) -> None:
        """Initialize Bag instance.

        Parameters
        ----------
        tokenizer : _Tokenizer
            A tokenizer instance from the :py:mod:`abydos.tokenizer` package
        intersection_type : str
            Specifies the intersection type, and set type as a result:
            See :ref:`intersection_type <intersection_type>` description in
            :py:class:`_TokenDistance` for details.
        **kwargs
            Arbitrary keyword arguments

        Other Parameters
        ----------------
        qval : int
            The length of each q-gram. Using this parameter and tokenizer=None
            will cause the instance to use the QGram tokenizer with this
            q value.
        metric : _Distance
            A string distance measure class for use in the ``soft`` and
            ``fuzzy`` variants.
        threshold : float
            A threshold value, similarities above which are counted as
            members of the intersection for the ``fuzzy`` variant.


        .. versionadded:: 0.4.0

        """
        if tokenizer is None:
            tokenizer = CharacterTokenizer()
        super(Bag, self).__init__(
            tokenizer=tokenizer, intersection_type=intersection_type, **kwargs
        )

    def dist_abs(self, src: str, tar: str, normalized: bool = False) -> float:
        """Return the bag distance between two strings.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison
        normalized : bool
            Normalizes to [0, 1] if True

        Returns
        -------
        int or float
            Bag distance

        Examples
        --------
        >>> cmp = Bag()
        >>> cmp.dist_abs('cat', 'hat')
        1
        >>> cmp.dist_abs('Niall', 'Neil')
        2
        >>> cmp.dist_abs('aluminum', 'Catalan')
        5
        >>> cmp.dist_abs('ATCG', 'TAGC')
        0
        >>> cmp.dist_abs('abcdefg', 'hijklm')
        7
        >>> cmp.dist_abs('abcdefg', 'hijklmno')
        8


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        if tar == src:
            return 0
        elif not src:
            return len(tar)
        elif not tar:
            return len(src)

        self._tokenize(src, tar)

        dist = max(self._src_only_card(), self._tar_only_card())

        if normalized:
            dist /= max(self._src_card(), self._tar_card())

        return dist

    def dist(self, src: str, tar: str) -> float:
        """Return the normalized bag distance between two strings.

        Bag distance is normalized by dividing by :math:`max( |src|, |tar| )`.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            Normalized bag distance

        Examples
        --------
        >>> cmp = Bag()
        >>> cmp.dist('cat', 'hat')
        0.3333333333333333
        >>> cmp.dist('Niall', 'Neil')
        0.4
        >>> cmp.dist('aluminum', 'Catalan')
        0.625
        >>> cmp.dist('ATCG', 'TAGC')
        0.0


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        if tar == src:
            return 0.0
        if not src or not tar:
            return 1.0

        return self.dist_abs(src, tar, normalized=True)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._lcsseq.

Longest common subsequence
"""

from typing import Any, Callable, List

from numpy import int_ as np_int
from numpy import zeros as np_zeros

from ._distance import _Distance

__all__ = ['LCSseq']


class LCSseq(_Distance):
    """Longest common subsequence.

    Longest common subsequence (LCSseq) is the longest subsequence of
    characters that two strings have in common.

    .. versionadded:: 0.3.6
    """

    def __init__(
        self, normalizer: Callable[[List[float]], float] = max, **kwargs: Any
    ) -> None:
        r"""Initialize LCSseq.

        Parameters
        ----------
        normalizer : function
            A normalization function for the normalized similarity & distance.
            By default, the max of the lengths of the input strings. If
            lambda x: sum(x)/2.0 is supplied, the normalization proposed in
            :cite:`Radev:2001` is used, i.e.
            :math:`\frac{2 \dot |LCS(src, tar)|}{|src| + |tar|}`.
        **kwargs
            Arbitrary keyword arguments


        .. versionadded:: 0.4.0

        """
        super(LCSseq, self).__init__(**kwargs)
        self._normalizer = normalizer

    def lcsseq(self, src: str, tar: str) -> str:
        """Return the longest common subsequence of two strings.

        Based on the dynamic programming algorithm from
        http://rosettacode.org/wiki/Longest_common_subsequence
        :cite:`rosettacode:2018b`. This is licensed GFDL 1.2.

        Modifications include:
            conversion to a numpy array in place of a list of lists

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        str
            The longest common subsequence

        Examples
        --------
        >>> sseq = LCSseq()
        >>> sseq.lcsseq('cat', 'hat')
        'at'
        >>> sseq.lcsseq('Niall', 'Neil')
        'Nil'
        >>> sseq.lcsseq('aluminum', 'Catalan')
        'aln'
        >>> sseq.lcsseq('ATCG', 'TAGC')
        'AC'


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        lengths = np_zeros((len(src) + 1, len(tar) + 1), dtype=np_int)

        # row 0 and column 0 are initialized to 0 already
        for i, src_char in enumerate(src):
            for j, tar_char in enumerate(tar):
                if src_char == tar_char:
                    lengths[i + 1, j + 1] = lengths[i, j] + 1
                else:
                    lengths[i + 1, j + 1] = max(
                        lengths[i + 1, j], lengths[i, j + 1]
                    )

        # read the substring out from the matrix
        result = ''
        i, j = len(src), len(tar)
        while i != 0 and j != 0:
            if lengths[i, j] == lengths[i - 1, j]:
                i -= 1
            elif lengths[i, j] == lengths[i, j - 1]:
                j -= 1
            else:
                result = src[i - 1] + result
                i -= 1
                j -= 1
        return result

    def sim(self, src: str, tar: str) -> float:
        r"""Return the longest common subsequence similarity of two strings.

        Longest common subsequence similarity (:math:`sim_{LCSseq}`).

        This employs the LCSseq function to derive a similarity metric:
        :math:`sim_{LCSseq}(s,t) = \frac{|LCSseq(s,t)|}{max(|s|, |t|)}`

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            LCSseq similarity

        Examples
        --------
        >>> sseq = LCSseq()
        >>> sseq.sim('cat', 'hat')
        0.6666666666666666
        >>> sseq.sim('Niall', 'Neil')
        0.6
        >>> sseq.sim('aluminum', 'Catalan')
        0.375
        >>> sseq.sim('ATCG', 'TAGC')
        0.5

        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class
        .. versionchanged:: 0.4.0
            Added normalization option

        """
        if src == tar:
            return 1.0
        elif not src or not tar:
            return 0.0
        return len(self.lcsseq(src, tar)) / self._normalizer(
            [len(src), len(tar)]
        )


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._ssk.

String subsequence kernel (SSK) similarity
"""

from typing import Any, Optional

from ._token_distance import _TokenDistance
from ._tokenizer import _Tokenizer
from ._q_skipgrams import QSkipgrams

__all__ = ['SSK']


class SSK(_TokenDistance):
    r"""String subsequence kernel (SSK) similarity.

    This is based on :cite:`Lodhi:2002`.


    .. versionadded:: 0.4.1
    """

    def __init__(
        self,
        tokenizer: Optional[_Tokenizer] = None,
        ssk_lambda: float = 0.9,
        **kwargs: Any
    ) -> None:
        """Initialize SSK instance.

        Parameters
        ----------
        tokenizer : _Tokenizer
            A tokenizer instance from the :py:mod:`abydos.tokenizer` package
        ssk_lambda : float or Iterable
            A value in the range (0.0, 1.0) used for discouting gaps between
            characters according to the method described in :cite:`Lodhi:2002`.
            To supply multiple values of lambda, provide an Iterable of numeric
            values, such as (0.5, 0.05) or np.arange(0.05, 0.5, 0.05)
        **kwargs
            Arbitrary keyword arguments

        Other Parameters
        ----------------
        qval : int
            The length of each q-skipgram. Using this parameter and
            tokenizer=None will cause the instance to use the QGramskipgrams
            tokenizer with this q value.


        .. versionadded:: 0.4.1

        """
        super(SSK, self).__init__(
            tokenizer=tokenizer, ssk_lambda=ssk_lambda, **kwargs
        )

        qval = 2 if 'qval' not in self.params else self.params['qval']
        self.params['tokenizer'] = (
            tokenizer
            if tokenizer is not None
            else QSkipgrams(
                qval=qval, start_stop='', scaler='SSK', ssk_lambda=ssk_lambda
            )
        )

    def sim_score(self, src: str, tar: str) -> float:
        """Return the SSK similarity of two strings.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            String subsequence kernel similarity

        Examples
        --------
        >>> cmp = SSK()
        >>> cmp.dist_abs('cat', 'hat')
        0.6441281138790036
        >>> cmp.dist_abs('Niall', 'Neil')
        0.5290992177869402
        >>> cmp.dist_abs('aluminum', 'Catalan')
        0.862398428061774
        >>> cmp.dist_abs('ATCG', 'TAGC')
        0.38591004719395017


        .. versionadded:: 0.4.1

        """
        self._tokenize(src, tar)

        src_wts = self._src_tokens
        tar_wts = self._tar_tokens

        score = sum(
            src_wts[token] * tar_wts[token] for token in src_wts & tar_wts
        )

        return score

    def sim(self, src: str, tar: str) -> float:
        """Return the normalized SSK similarity of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Normalized string subsequence kernel similarity

        Examples
        --------
        >>> cmp = SSK()
        >>> cmp.sim('cat', 'hat')
        0.3558718861209964
        >>> cmp.sim('Niall', 'Neil')
        0.4709007822130597
        >>> cmp.sim('aluminum', 'Catalan')
        0.13760157193822603
        >>> cmp.sim('ATCG', 'TAGC')
        0.6140899528060498


        .. versionadded:: 0.4.1

        """
        if src == tar:
            return 1.0

        self._tokenize(src, tar)

        src_wts = self._src_tokens
        tar_wts = self._tar_tokens

        score = sum(
            src_wts[token] * tar_wts[token] for token in src_wts & tar_wts
        )

        norm = (
            sum(src_wts[token] * src_wts[token] for token in src_wts)
            * sum(tar_wts[token] * tar_wts[token] for token in tar_wts)
        ) ** 0.5

        if not score:
            return 0.0
        return score / norm


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

from ._bag import Bag
from ._baulieu_xiii import BaulieuXIII
from ._clement import Clement
from ._cormode_lz import CormodeLZ
from ._damerau_levenshtein import DamerauLevenshtein
from ._dice_asymmetric_i import DiceAsymmetricI
from ._discounted_levenshtein import DiscountedLevenshtein
from ._distance import _Distance
from ._editex import Editex
from ._fuzzywuzzy_partial_string import FuzzyWuzzyPartialString
from ._fuzzywuzzy_token_set import FuzzyWuzzyTokenSet
from ._fuzzywuzzy_token_sort import FuzzyWuzzyTokenSort
from ._hamming import Hamming
from ._indel import Indel
from ._iterative_substring import IterativeSubString
from ._kuhns_iii import KuhnsIII
from ._lcprefix import LCPrefix
from ._lcsseq import LCSseq
from ._levenshtein import Levenshtein
from ._lig3 import LIG3
from ._ncd_bz2 import NCDbz2
from ._overlap import Overlap
from ._pearson_chi_squared import PearsonChiSquared
from ._pearson_ii import PearsonII
from ._ratcliff_obershelp import RatcliffObershelp
from ._rouge_l import RougeL
from ._ssk import SSK
from ._tichy import Tichy
from ._phonetic_distance import PhoneticDistance
from ._double_metaphone import DoubleMetaphone
from ._refined_soundex import RefinedSoundex
from ._token_distance import _TokenDistance
from ._typo import Typo
from ._warrens_iv import WarrensIV
from ._weighted_jaccard import WeightedJaccard
from ._character import CharacterTokenizer
from ._q_grams import QGrams
from ._q_skipgrams import QSkipgrams
from ._regexp import RegexpTokenizer
from ._tokenizer import _Tokenizer
from ._whitespace import WhitespaceTokenizer


# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._phonetic_distance.

Phonetic distance.
"""

from typing import Any, Callable, List, Optional, Sequence, Type, Union

from ._distance import _Distance
from ._phonetic import _Phonetic

__all__ = ['PhoneticDistance']


class PhoneticDistance(_Distance):
    """Phonetic distance.

    Phonetic distance applies one or more supplied string transformations to
    words and compares the resulting transformed strings using a supplied
    distance measure.

    A simple example would be to create a 'Soundex distance':

    >>> from abydos.phonetic import Soundex
    >>> soundex = PhoneticDistance(transforms=Soundex())
    >>> soundex.dist('Ashcraft', 'Ashcroft')
    0.0
    >>> soundex.dist('Robert', 'Ashcraft')
    1.0

    .. versionadded:: 0.4.1
    """

    def __init__(
        self,
        transforms: Optional[
            Union[
                Type[_Phonetic],
                _Phonetic,
                Callable[[str], str],
                Sequence[
                    Union[
                        Type[_Phonetic],
                        _Phonetic,
                        Callable[[str], str],
                    ]
                ],
            ]
        ] = None,
        metric: Optional[Union[Type[_Distance], _Distance]] = None,
        encode_alpha: bool = False,
        **kwargs: Any
    ) -> None:
        """Initialize PhoneticDistance instance.

        Parameters
        ----------
        transforms : list or _Phonetic or _Stemmer or _Fingerprint or type
            An instance of a subclass of _Phonetic, _Stemmer, or _Fingerprint,
            or a list (or other iterable) of such instances to apply to each
            input word before computing their distance or similarity. If
            omitted, no transformations will be performed.
        metric : _Distance or type
            An instance of a subclass of _Distance, used for computing the
            inputs' distance or similarity after being transformed. If omitted,
            the strings will be compared for identify (returning 0.0 if
            identical, otherwise 1.0, when distance is computed).
        encode_alpha : bool
            Set to true to use the encode_alpha method of phonetic algorithms
            whenever possible.
        **kwargs
            Arbitrary keyword arguments


        .. versionadded:: 0.4.1

        """
        super(PhoneticDistance, self).__init__(**kwargs)
        self.transforms = []  # type: List[Callable[[str], str]]
        if transforms:
            if isinstance(transforms, Sequence):
                transforms = list(transforms)
            else:
                transforms = [transforms]

            for i, trans in enumerate(transforms):
                if isinstance(trans, (_Phonetic)):
                    continue
                elif isinstance(trans, type) and issubclass(
                    trans, (_Phonetic)
                ):
                    transforms[i] = trans()
                elif callable(trans):
                    continue
                else:
                    raise TypeError(
                        '{} has unknown type {}'.format(trans, type(trans))
                    )

            for trans in transforms:
                if isinstance(trans, _Phonetic):
                    if encode_alpha:
                        self.transforms.append(trans.encode_alpha)
                    else:
                        self.transforms.append(trans.encode)
                else:  # callable(trans)
                    self.transforms.append(trans)  # type: ignore

        if isinstance(metric, type) and issubclass(metric, _Distance):
            self.metric = metric()  # type: Optional[_Distance]
        elif isinstance(metric, _Distance):
            self.metric = metric
        elif metric is None:
            self.metric = None
        else:
            raise TypeError(
                '{} has unknown type {}'.format(metric, type(metric))
            )

    def dist_abs(self, src: str, tar: str) -> float:
        """Return the Phonetic distance.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float or int
            The Phonetic distance

        Examples
        --------
        >>> from abydos.phonetic import Soundex
        >>> cmp = PhoneticDistance(Soundex())
        >>> cmp.dist_abs('cat', 'hat')
        1
        >>> cmp.dist_abs('Niall', 'Neil')
        0
        >>> cmp.dist_abs('Colin', 'Cuilen')
        0
        >>> cmp.dist_abs('ATCG', 'TAGC')
        1

        >>> from abydos.distance import Levenshtein
        >>> cmp = PhoneticDistance(transforms=[Soundex], metric=Levenshtein)
        >>> cmp.dist_abs('cat', 'hat')
        1
        >>> cmp.dist_abs('Niall', 'Neil')
        0
        >>> cmp.dist_abs('Colin', 'Cuilen')
        0
        >>> cmp.dist_abs('ATCG', 'TAGC')
        3


        .. versionadded:: 0.4.1

        """
        for trans in self.transforms:
            src = trans(src)
            tar = trans(tar)
        if self.metric:
            return self.metric.dist_abs(src, tar)
        else:
            return int(src != tar)

    def dist(self, src: str, tar: str) -> float:
        """Return the normalized Phonetic distance.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            The normalized Phonetic distance

        Examples
        --------
        >>> from abydos.phonetic import Soundex
        >>> cmp = PhoneticDistance(Soundex())
        >>> cmp.dist('cat', 'hat')
        1.0
        >>> cmp.dist('Niall', 'Neil')
        0.0
        >>> cmp.dist('Colin', 'Cuilen')
        0.0
        >>> cmp.dist('ATCG', 'TAGC')
        1.0

        >>> from abydos.distance import Levenshtein
        >>> cmp = PhoneticDistance(transforms=[Soundex], metric=Levenshtein)
        >>> cmp.dist('cat', 'hat')
        0.25
        >>> cmp.dist('Niall', 'Neil')
        0.0
        >>> cmp.dist('Colin', 'Cuilen')
        0.0
        >>> cmp.dist('ATCG', 'TAGC')
        0.75


        .. versionadded:: 0.4.1

        """
        for trans in self.transforms:
            src = trans(src)
            tar = trans(tar)
        if self.metric:
            return self.metric.dist(src, tar)
        else:
            return float(src != tar)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._ratcliff_obershelp.

Ratcliff-Obershelp similarity
"""

from typing import Tuple

from numpy import int_ as np_int
from numpy import zeros as np_zeros

from ._distance import _Distance

__all__ = ['RatcliffObershelp']


class RatcliffObershelp(_Distance):
    """Ratcliff-Obershelp similarity.

    This follows the Ratcliff-Obershelp algorithm :cite:`Ratcliff:1988` to
    derive a similarity measure:

        1. Find the length of the longest common substring in src & tar.
        2. Recurse on the strings to the left & right of each this substring
           in src & tar. The base case is a 0 length common substring, in which
           case, return 0. Otherwise, return the sum of the current longest
           common substring and the left & right recursed sums.
        3. Multiply this length by 2 and divide by the sum of the lengths of
           src & tar.

    Cf.
    http://www.drdobbs.com/database/pattern-matching-the-gestalt-approach/184407970

    .. versionadded:: 0.3.6
    """

    def sim(self, src: str, tar: str) -> float:
        """Return the Ratcliff-Obershelp similarity of two strings.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            Ratcliff-Obershelp similarity

        Examples
        --------
        >>> cmp = RatcliffObershelp()
        >>> round(cmp.sim('cat', 'hat'), 12)
        0.666666666667
        >>> round(cmp.sim('Niall', 'Neil'), 12)
        0.666666666667
        >>> round(cmp.sim('aluminum', 'Catalan'), 12)
        0.4
        >>> cmp.sim('ATCG', 'TAGC')
        0.5


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """

        def _lcsstr_stl(src: str, tar: str) -> Tuple[int, int, int]:
            """Return start positions & length for Ratcliff-Obershelp.

            Parameters
            ----------
            src : str
                Source string for comparison
            tar : str
            Target string for comparison

            Returns
            -------
            tuple
                The start position in the source string, start position in the
                target string, and length of the longest common substring of
                strings src and tar.

            .. versionadded:: 0.1.0

            """
            lengths = np_zeros((len(src) + 1, len(tar) + 1), dtype=np_int)
            longest, src_longest, tar_longest = 0, 0, 0
            for i in range(1, len(src) + 1):
                for j in range(1, len(tar) + 1):
                    if src[i - 1] == tar[j - 1]:
                        lengths[i, j] = lengths[i - 1, j - 1] + 1
                        if lengths[i, j] > longest:
                            longest = lengths[i, j]
                            src_longest = i
                            tar_longest = j
                    else:
                        lengths[i, j] = 0
            return src_longest - longest, tar_longest - longest, longest

        def _sstr_matches(src: str, tar: str) -> int:
            """Return the sum of substring match lengths.

            This follows the Ratcliff-Obershelp algorithm
            :cite:`Ratcliff:1988`:
                 1. Find the length of the longest common substring in src &
                     tar.
                 2. Recurse on the strings to the left & right of each this
                     substring in src & tar.
                 3. Base case is a 0 length common substring, in which case,
                     return 0.
                 4. Return the sum.

            Parameters
            ----------
            src : str
                Source string for comparison
            tar : str
                Target string for comparison

            Returns
            -------
            int
                Sum of substring match lengths

            .. versionadded:: 0.1.0

            """
            src_start, tar_start, length = _lcsstr_stl(src, tar)
            if length == 0:
                return 0
            return (
                _sstr_matches(src[:src_start], tar[:tar_start])
                + length
                + _sstr_matches(
                    src[src_start + length :], tar[tar_start + length :]
                )
            )

        if src == tar:
            return 1.0
        elif not src or not tar:
            return 0.0
        return 2 * _sstr_matches(src, tar) / (len(src) + len(tar))


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._fuzzywuzzy_partial_string.

FuzzyWuzzy Partial String similarity
"""

from difflib import SequenceMatcher

from ._distance import _Distance

__all__ = ['FuzzyWuzzyPartialString']


class FuzzyWuzzyPartialString(_Distance):
    """FuzzyWuzzy Partial String similarity.

    This follows the FuzzyWuzzy Partial String similarity algorithm
    :cite:`Cohen:2011`. Rather than returning an integer in the range [0, 100],
    as demonstrated in the blog post, this implementation returns a float in
    the range [0.0, 1.0].

    .. versionadded:: 0.4.0
    """

    def sim(self, src: str, tar: str) -> float:
        """Return the FuzzyWuzzy Partial String similarity of two strings.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            FuzzyWuzzy Partial String similarity

        Examples
        --------
        >>> cmp = FuzzyWuzzyPartialString()
        >>> round(cmp.sim('cat', 'hat'), 12)
        0.666666666667
        >>> round(cmp.sim('Niall', 'Neil'), 12)
        0.75
        >>> round(cmp.sim('aluminum', 'Catalan'), 12)
        0.428571428571
        >>> cmp.sim('ATCG', 'TAGC')
        0.5


        .. versionadded:: 0.4.0

        """
        max_sim = 0.0
        start_pos = 0

        if len(src) > len(tar):
            src, tar = tar, src

        src_len = len(src)

        while max_sim < 1.0 and start_pos < len(tar) - src_len + 1:
            max_sim = max(
                max_sim,
                SequenceMatcher(
                    None, src, tar[start_pos : start_pos + src_len]
                ).ratio(),
            )
            start_pos += 1

        return max_sim


if __name__ == '__main__':
    import doctest

    doctest.testmod()



import numpy as np
from tqdm import tqdm
# from numba import jit
from scipy.sparse import csc_matrix, coo_matrix
from typing import Union

# @jit(nopython=True, fastmath=True)
def _sparse_cosine_low_memory(matrix_row: np.array,
                              matrix_col: np.array,
                              matrix_data: np.array,
                              matrix_len: int,
                              vector_ind: np.array,
                              vector_data: np.array) -> np.array:
    """
    A sparse cosine simularity calculation between a matrix and a vector. The sparse matrix should be sorted
    in ascending order based on the matrix_col values. The vector should be sorted based on the indexes in 
    ascending order.

    Parameters
    ----------
    matrix_row : np.array
        The row indices of the ngrams matrix of the matching data
    matrix_col : np.array
        The column indices of the ngrams matrix of the matching data in ascending order
    matrix_data : np.array
        The data of the ngrams matrix of the matching data
    matrix_len : int
        The length (number of rows) of the ngrams matrix of the matching data
    vector_ind : np.array
        The indices of the ngrams vector of the to be matched data
    vector_data : np.array
        The data of the ngrams vector of the to be matched data

    Returns
    -------
    np.array
        The cosine simularity between each of the rows of the matrix and the vector

    """
    ind = 0
    res = np.zeros(matrix_len, np.float32)
    for mat_ind in range(len(matrix_col)):
        col = matrix_col[mat_ind]
        if col > vector_ind[ind]:
            ind = ind + 1
            if ind == len(vector_ind):
                break
        if col == vector_ind[ind]:
            res[matrix_row[mat_ind]] = res[matrix_row[mat_ind]] + \
                matrix_data[mat_ind] * vector_data[ind]

    return res


def _sparse_cosine_top_n_standard(matrix_a: csc_matrix,
                                  matrix_b: csc_matrix,
                                  number_of_rows_at_once: int,
                                  top_n: int,
                                  verbose: bool) -> np.array:
    """
    A function for sparse matrix multiplication followed by an argpartition to
    only take the top_n indexes.

    Parameters
    -------
    matrix_a: csc_matric
        The largest sparse csc matrix which should be multiplied
    matrix_b: csc_matric
        The smallest sparse csc matrix which should be multiplied
    number_of_rows_at_once: int
        The number of rows which should be processed at once, a lower
        number of rows reduces the memory usage
    top_n: int
        The best n matches that should be returned
    verbose: bool
        A boolean indicating whether the progress should be printed

    Returns
    -------
    np.array
        The indexes for the n best sparse cosine matches between matrix a and b

    """

    results_arg = np.zeros(
        (matrix_b.shape[0], top_n), dtype=np.float32)

    # Split up the matrice in a certain number of rows
    for j in tqdm(range(0, matrix_b.shape[0], number_of_rows_at_once), disable=not verbose):
        number_of_rows_at_once_min = min(
            [number_of_rows_at_once, matrix_b.shape[0]-j])
        matrix_b_temp = matrix_b[j:j+number_of_rows_at_once_min, :]

        # Calculate the matrix dot product
        results_full = (matrix_a * (matrix_b_temp.T)).tocsc()

        # For each of the rows of the original matrix select the argpartition
        for i in range(number_of_rows_at_once_min):
            results_full_temp = results_full.data[results_full.indptr[i]:results_full.indptr[i+1]]

            # If there are more results then top_n only select the top_n results
            if len(results_full_temp) > top_n:
                ind = results_full.indices[results_full.indptr[i]:results_full.indptr[i+1]]
                results_arg[j + i, :] = ind[np.argpartition(
                    results_full_temp, -top_n)[-top_n:]]
            
            # else just select all the results
            else:
                results_arg[j + i, :len(results_full_temp)
                            ] = results_full.indices[results_full.indptr[i]:results_full.indptr[i+1]]
    return results_arg

def sparse_cosine_top_n(matrix_a: Union[csc_matrix, coo_matrix], 
                        matrix_b: csc_matrix, 
                        top_n: int, 
                        low_memory: bool,
                        number_of_rows: int,
                        verbose: bool):
    """
    Calculates the top_n cosine matches between matrix_a and matrix_b. Takes into account
    the amount of  memory that should be used based on the low_memory int

    Parameters
    -------
    matrix_a: csc_matric
        The largest sparse csc matrix which should be multiplied
    matrix_b: csc_matric
        The smallest sparse csc matrix which should be multiplied
    top_n: int
        The best n matches that should be returned
    low_memory: bool
        A bool indicating whether the low memory sparse cosine approach should be used
    number_of_rows: int
        An int inidcating the number of rows which should be 
        processed at once when calculating the cosine simalarity
    verbose: bool
        A boolean indicating whether the progress should be printed

    Returns
    -------
    np.array
        The indexes for the n best sparse cosine matches between matrix a and b

    """
    if low_memory:
        matrix_b.sort_indices()
        res = _sparse_cosine_low_memory(matrix_a.row, matrix_a.col, matrix_a.data,
                        matrix_a.shape[0], matrix_b.indices, matrix_b.data)

        top_n_adjusted = -np.min([top_n, len(res)])

        return np.argpartition(res, top_n_adjusted, axis=0)[top_n_adjusted:]
    else:
        return _sparse_cosine_top_n_standard(matrix_a, matrix_b, number_of_rows, top_n, verbose)


# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._pearson_chi_squared.

Pearson's Chi-Squared similarity
"""

from math import copysign
from typing import Any, Counter as TCounter, Optional, Sequence, Set, Union

from ._token_distance import _TokenDistance
from ._tokenizer import _Tokenizer

__all__ = ['PearsonChiSquared']


class PearsonChiSquared(_TokenDistance):
    r"""Pearson's Chi-Squared similarity.

    For two sets X and Y and a population N, the Pearson's :math:`\chi^2`
    similarity :cite:`Pearson:1913` is

        .. math::

            sim_{PearsonChiSquared}(X, Y) =
            \frac{|N| \cdot (|X \cap Y| \cdot |(N \setminus X) \setminus Y| -
            |X \setminus Y| \cdot |Y \setminus X|)^2}
            {|X| \cdot |Y| \cdot |N \setminus X| \cdot |N \setminus Y|}

    This is also Pearson I similarity.

    In :ref:`2x2 confusion table terms <confusion_table>`, where a+b+c+d=n,
    this is

        .. math::

            sim_{PearsonChiSquared} =
            \frac{n(ad-bc)^2}{(a+b)(a+c)(b+d)(c+d)}

    .. versionadded:: 0.4.0
    """

    def __init__(
        self,
        alphabet: Optional[
            Union[TCounter[str], Sequence[str], Set[str], int]
        ] = None,
        tokenizer: Optional[_Tokenizer] = None,
        intersection_type: str = 'crisp',
        **kwargs: Any
    ) -> None:
        """Initialize PearsonChiSquared instance.

        Parameters
        ----------
        alphabet : Counter, collection, int, or None
            This represents the alphabet of possible tokens.
            See :ref:`alphabet <alphabet>` description in
            :py:class:`_TokenDistance` for details.
        tokenizer : _Tokenizer
            A tokenizer instance from the :py:mod:`abydos.tokenizer` package
        intersection_type : str
            Specifies the intersection type, and set type as a result:
            See :ref:`intersection_type <intersection_type>` description in
            :py:class:`_TokenDistance` for details.
        **kwargs
            Arbitrary keyword arguments

        Other Parameters
        ----------------
        qval : int
            The length of each q-gram. Using this parameter and tokenizer=None
            will cause the instance to use the QGram tokenizer with this
            q value.
        metric : _Distance
            A string distance measure class for use in the ``soft`` and
            ``fuzzy`` variants.
        threshold : float
            A threshold value, similarities above which are counted as
            members of the intersection for the ``fuzzy`` variant.


        .. versionadded:: 0.4.0

        """
        super(PearsonChiSquared, self).__init__(
            alphabet=alphabet,
            tokenizer=tokenizer,
            intersection_type=intersection_type,
            **kwargs
        )

    def sim_score(self, src: str, tar: str) -> float:
        """Return Pearson's Chi-Squared similarity of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Pearson's Chi-Squared similarity

        Examples
        --------
        >>> cmp = PearsonChiSquared()
        >>> cmp.sim_score('cat', 'hat')
        193.99489809335964
        >>> cmp.sim_score('Niall', 'Neil')
        101.99771068526542
        >>> cmp.sim_score('aluminum', 'Catalan')
        9.19249664336649
        >>> cmp.sim_score('ATCG', 'TAGC')
        0.032298410951138765


        .. versionadded:: 0.4.0

        """
        self._tokenize(src, tar)

        a = self._intersection_card()
        b = self._src_only_card()
        c = self._tar_only_card()
        d = self._total_complement_card()
        n = self._population_unique_card()
        ab = self._src_card()
        ac = self._tar_card()

        if src == tar:
            return float(n)
        if not src or not tar:
            return 0.0
        num = n * (a * d - b * c) ** 2
        if num:
            return num / (ab * ac * (b + d) * (c + d))
        return 0.0

    def corr(self, src: str, tar: str) -> float:
        """Return Pearson's Chi-Squared correlation of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Pearson's Chi-Squared correlation

        Examples
        --------
        >>> cmp = PearsonChiSquared()
        >>> cmp.corr('cat', 'hat')
        0.2474424720578567
        >>> cmp.corr('Niall', 'Neil')
        0.1300991207720222
        >>> cmp.corr('aluminum', 'Catalan')
        0.011710186806836291
        >>> cmp.corr('ATCG', 'TAGC')
        -4.1196952743799446e-05


        .. versionadded:: 0.4.0

        """
        if src == tar:
            return 1.0

        score = self.sim_score(src, tar)

        a = self._intersection_card()
        b = self._src_only_card()
        c = self._tar_only_card()
        d = self._total_complement_card()

        score /= a + b + c + d

        return copysign(score, a * d - b * c)

    def sim(self, src: str, tar: str) -> float:
        """Return Pearson's normalized Chi-Squared similarity of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Normalized Pearson's Chi-Squared similarity

        Examples
        --------
        >>> cmp = PearsonChiSquared()
        >>> cmp.corr('cat', 'hat')
        0.2474424720578567
        >>> cmp.corr('Niall', 'Neil')
        0.1300991207720222
        >>> cmp.corr('aluminum', 'Catalan')
        0.011710186806836291
        >>> cmp.corr('ATCG', 'TAGC')
        -4.1196952743799446e-05


        .. versionadded:: 0.4.0

        """
        return (1.0 + self.corr(src, tar)) / 2.0


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._distance.

The distance._distance module implements abstract class _Distance.
"""
from typing import Any, Dict

__all__ = ['_Distance']


class _Distance:
    """Abstract Distance class.

    .. versionadded:: 0.3.6
    """

    def __init__(self, **kwargs: Any) -> None:
        """Initialize _Distance instance.

        Parameters
        ----------
        **kwargs
            Arbitrary keyword arguments


        .. versionadded:: 0.4.0

        """
        self.params = {}  # type: Dict[str, Any]
        self.set_params(**kwargs)

    def set_params(self, **kwargs: Any) -> None:
        """Store params in the params dict.

        Parameters
        ----------
        **kwargs
            Arbitrary keyword arguments


        .. versionadded:: 0.4.0

        """
        for key in kwargs:
            self.params[key] = kwargs[key]

    def sim(self, src: str, tar: str) -> float:
        """Return similarity.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            Similarity


        .. versionadded:: 0.3.6

        """
        return 1.0 - self.dist(src, tar)

    def dist(self, src: str, tar: str) -> float:
        """Return distance.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            Distance


        .. versionadded:: 0.3.6

        """
        return 1.0 - self.sim(src, tar)

    def dist_abs(self, src: str, tar: str) -> float:
        """Return absolute distance.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        int
            Absolute distance


        .. versionadded:: 0.3.6

        """
        return self.dist(src, tar)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._dice_asymmetric_i.

Dice's Asymmetric I similarity
"""

from typing import Any, Optional

from ._token_distance import _TokenDistance
from ._tokenizer import _Tokenizer

__all__ = ['DiceAsymmetricI']


class DiceAsymmetricI(_TokenDistance):
    r"""Dice's Asymmetric I similarity.

    For two sets X and Y and a population N, Dice's Asymmetric I similarity
    :cite:`Dice:1945` is

        .. math::

            sim_{DiceAsymmetricI}(X, Y) =
            \frac{|X \cap Y|}{|X|}

    In :ref:`2x2 confusion table terms <confusion_table>`, where a+b+c+d=n,
    this is

        .. math::

            sim_{DiceAsymmetricI} =
            \frac{a}{a+b}

    Notes
    -----
    In terms of a confusion matrix, this is equivalent to precision or
    positive predictive value :py:meth:`ConfusionTable.precision`.

    .. versionadded:: 0.4.0

    """

    def __init__(
        self,
        tokenizer: Optional[_Tokenizer] = None,
        intersection_type: str = 'crisp',
        **kwargs: Any
    ) -> None:
        """Initialize DiceAsymmetricI instance.

        Parameters
        ----------
        tokenizer : _Tokenizer
            A tokenizer instance from the :py:mod:`abydos.tokenizer` package
        intersection_type : str
            Specifies the intersection type, and set type as a result:
            See :ref:`intersection_type <intersection_type>` description in
            :py:class:`_TokenDistance` for details.
        **kwargs
            Arbitrary keyword arguments

        Other Parameters
        ----------------
        qval : int
            The length of each q-gram. Using this parameter and tokenizer=None
            will cause the instance to use the QGram tokenizer with this
            q value.
        metric : _Distance
            A string distance measure class for use in the ``soft`` and
            ``fuzzy`` variants.
        threshold : float
            A threshold value, similarities above which are counted as
            members of the intersection for the ``fuzzy`` variant.


        .. versionadded:: 0.4.0

        """
        super(DiceAsymmetricI, self).__init__(
            tokenizer=tokenizer, intersection_type=intersection_type, **kwargs
        )

    def sim(self, src: str, tar: str) -> float:
        """Return the Dice's Asymmetric I similarity of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Dice's Asymmetric I similarity

        Examples
        --------
        >>> cmp = DiceAsymmetricI()
        >>> cmp.sim('cat', 'hat')
        0.5
        >>> cmp.sim('Niall', 'Neil')
        0.3333333333333333
        >>> cmp.sim('aluminum', 'Catalan')
        0.1111111111111111
        >>> cmp.sim('ATCG', 'TAGC')
        0.0


        .. versionadded:: 0.4.0

        """
        if src == tar:
            return 1.0

        self._tokenize(src, tar)

        a = self._intersection_card()
        ab = self._src_card()

        if a == 0.0:
            return 0.0
        return a / ab


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._discounted_levenshtein.

Discounted Levenshtein edit distance
"""

from math import log
from typing import Any, Callable, List, Tuple, Union, cast

import numpy as np

from ._levenshtein import Levenshtein

__all__ = ['DiscountedLevenshtein']


class DiscountedLevenshtein(Levenshtein):
    """Discounted Levenshtein distance.

    This is a variant of Levenshtein distance for which edits later in a string
    have discounted cost, on the theory that earlier edits are less likely
    than later ones.

    .. versionadded:: 0.4.1
    """

    def __init__(
        self,
        mode: str = 'lev',
        normalizer: Callable[[List[float]], float] = max,
        discount_from: Union[int, str] = 1,
        discount_func: Union[str, Callable[[float], float]] = 'log',
        vowels: str = 'aeiou',
        **kwargs: Any
    ) -> None:
        """Initialize DiscountedLevenshtein instance.

        Parameters
        ----------
        mode : str
            Specifies a mode for computing the discounted Levenshtein distance:

                - ``lev`` (default) computes the ordinary Levenshtein distance,
                  in which edits may include inserts, deletes, and
                  substitutions
                - ``osa`` computes the Optimal String Alignment distance, in
                  which edits may include inserts, deletes, substitutions, and
                  transpositions but substrings may only be edited once

        normalizer : function
            A function that takes an list and computes a normalization term
            by which the edit distance is divided (max by default). Another
            good option is the sum function.
        discount_from : int or str
            If an int is supplied, this is the first character whose edit cost
            will be discounted. If the str ``coda`` is supplied, discounting
            will start with the first non-vowel after the first vowel (the
            first syllable coda).
        discount_func : str or function
            The two supported str arguments are ``log``, for a logarithmic
            discount function, and ``exp`` for a exponential discount function.
            See notes below for information on how to supply your own
            discount function.
        vowels : str
            These are the letters to consider as vowels when discount_from is
            set to ``coda``. It defaults to the English vowels 'aeiou', but
            it would be reasonable to localize this to other languages or to
            add orthographic semi-vowels like 'y', 'w', and even 'h'.
        **kwargs
            Arbitrary keyword arguments

        Notes
        -----
        This class is highly experimental and will need additional tuning.

        The discount function can be passed as a callable function. It should
        expect an integer as its only argument and return a float, ideally
        less than or equal to 1.0. The argument represents the degree of
        discounting to apply.


        .. versionadded:: 0.4.1

        """
        super(DiscountedLevenshtein, self).__init__(**kwargs)
        self._mode = mode
        self._normalizer = normalizer
        self._discount_from = discount_from
        self._vowels = set(vowels.lower())
        if callable(discount_func):
            self._discount_func = discount_func
        elif discount_func == 'exp':
            self._discount_func = self._exp_discount
        else:
            self._discount_func = self._log_discount

    @staticmethod
    def _log_discount(discounts: float) -> float:
        return 1 / (log(1 + discounts / 5) + 1)

    @staticmethod
    def _exp_discount(discounts: float) -> float:
        return 1 / (discounts + 1) ** 0.2

    def _alignment_matrix(
        self, src: str, tar: str, backtrace: bool = True
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Return the Levenshtein alignment matrix.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison
        backtrace : bool
            Return the backtrace matrix as well

        Returns
        -------
        numpy.ndarray or tuple(numpy.ndarray, numpy.ndarray)
            The alignment matrix and (optionally) the backtrace matrix


        .. versionadded:: 0.4.1

        """
        src_len = len(src)
        tar_len = len(tar)

        if self._discount_from == 'coda':
            discount_from = [0, 0]

            src_voc = src.lower()
            for i in range(len(src_voc)):
                if src_voc[i] in self._vowels:
                    discount_from[0] = i
                    break
            for i in range(discount_from[0], len(src_voc)):
                if src_voc[i] not in self._vowels:
                    discount_from[0] = i
                    break
            else:
                discount_from[0] += 1

            tar_voc = tar.lower()
            for i in range(len(tar_voc)):
                if tar_voc[i] in self._vowels:
                    discount_from[1] = i
                    break
            for i in range(discount_from[1], len(tar_voc)):
                if tar_voc[i] not in self._vowels:
                    discount_from[1] = i
                    break
            else:
                discount_from[1] += 1

        elif isinstance(self._discount_from, int):
            discount_from = [self._discount_from, self._discount_from]
        else:
            discount_from = [1, 1]

        d_mat = np.zeros((src_len + 1, tar_len + 1), dtype=np.float_)
        if backtrace:
            trace_mat = np.zeros((src_len + 1, tar_len + 1), dtype=np.int8)
        for i in range(1, src_len + 1):
            d_mat[i, 0] = d_mat[i - 1, 0] + self._discount_func(
                max(0, i - discount_from[0])
            )
            if backtrace:
                trace_mat[i, 0] = 1
        for j in range(1, tar_len + 1):
            d_mat[0, j] = d_mat[0, j - 1] + self._discount_func(
                max(0, j - discount_from[1])
            )
            if backtrace:
                trace_mat[0, j] = 0
        for i in range(src_len):
            i_extend = self._discount_func(max(0, i - discount_from[0]))
            for j in range(tar_len):
                traces = ((i + 1, j), (i, j + 1), (i, j))
                cost = min(
                    i_extend, self._discount_func(max(0, j - discount_from[1]))
                )
                opts = (
                    d_mat[traces[0]] + cost,  # ins
                    d_mat[traces[1]] + cost,  # del
                    d_mat[traces[2]]
                    + (cost if src[i] != tar[j] else 0),  # sub/==
                )
                d_mat[i + 1, j + 1] = min(opts)
                if backtrace:
                    trace_mat[i + 1, j + 1] = int(np.argmin(opts))

                if self._mode == 'osa':
                    if (
                        i + 1 > 1
                        and j + 1 > 1
                        and src[i] == tar[j - 1]
                        and src[i - 1] == tar[j]
                    ):
                        # transposition
                        d_mat[i + 1, j + 1] = min(
                            d_mat[i + 1, j + 1], d_mat[i - 1, j - 1] + cost
                        )
                        if backtrace:
                            trace_mat[i + 1, j + 1] = 2
        if backtrace:
            return d_mat, trace_mat
        return d_mat

    def dist_abs(self, src: str, tar: str) -> float:
        """Return the Levenshtein distance between two strings.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float (may return a float if cost has float values)
            The Levenshtein distance between src & tar

        Examples
        --------
        >>> cmp = DiscountedLevenshtein()
        >>> cmp.dist_abs('cat', 'hat')
        1
        >>> cmp.dist_abs('Niall', 'Neil')
        2.526064024369237
        >>> cmp.dist_abs('aluminum', 'Catalan')
        5.053867269967515
        >>> cmp.dist_abs('ATCG', 'TAGC')
        2.594032108779918

        >>> cmp = DiscountedLevenshtein(mode='osa')
        >>> cmp.dist_abs('ATCG', 'TAGC')
        1.7482385137517997
        >>> cmp.dist_abs('ACTG', 'TAGC')
        3.342270622531718


        .. versionadded:: 0.4.1

        """
        src_len = len(src)
        tar_len = len(tar)

        if src == tar:
            return 0.0

        if isinstance(self._discount_from, int):
            discount_from = self._discount_from
        else:
            discount_from = 1

        if not src:
            return sum(
                self._discount_func(max(0, pos - discount_from))
                for pos in range(tar_len)
            )
        if not tar:
            return sum(
                self._discount_func(max(0, pos - discount_from))
                for pos in range(src_len)
            )

        d_mat = cast(
            np.ndarray, self._alignment_matrix(src, tar, backtrace=False)
        )

        if int(d_mat[src_len, tar_len]) == d_mat[src_len, tar_len]:
            return int(d_mat[src_len, tar_len])
        else:
            return cast(float, d_mat[src_len, tar_len])

    def dist(self, src: str, tar: str) -> float:
        """Return the normalized Levenshtein distance between two strings.

        The Levenshtein distance is normalized by dividing the Levenshtein
        distance (calculated by any of the three supported methods) by the
        greater of the number of characters in src times the cost of a delete
        and the number of characters in tar times the cost of an insert.
        For the case in which all operations have :math:`cost = 1`, this is
        equivalent to the greater of the length of the two strings src & tar.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            The normalized Levenshtein distance between src & tar

        Examples
        --------
        >>> cmp = DiscountedLevenshtein()
        >>> cmp.dist('cat', 'hat')
        0.3513958291799864
        >>> cmp.dist('Niall', 'Neil')
        0.5909885886270658
        >>> cmp.dist('aluminum', 'Catalan')
        0.8348163322045603
        >>> cmp.dist('ATCG', 'TAGC')
        0.7217609721523955


        .. versionadded:: 0.4.1

        """
        if src == tar:
            return 0

        if isinstance(self._discount_from, int):
            discount_from = self._discount_from
        else:
            discount_from = 1

        src_len = len(src)
        tar_len = len(tar)

        normalize_term = self._normalizer(
            [
                sum(
                    self._discount_func(max(0, pos - discount_from))
                    for pos in range(src_len)
                ),
                sum(
                    self._discount_func(max(0, pos - discount_from))
                    for pos in range(tar_len)
                ),
            ]
        )

        return self.dist_abs(src, tar) / normalize_term


if __name__ == '__main__':
    import doctest

    doctest.testmod()



from setuptools import setup
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
   name='name_matching',
   version='0.8.7',
   description='A package for the matching of company names',
   author='Michiel Nijhuis',
   author_email='m.nijhuis@dnb.nl',
   project_urls = {
        'Documentation': 'https://name-matching.readthedocs.io/en/latest/index.html',
        'Source Code': 'https://github.com/DeNederlandscheBank/name_matching'},
   packages=['name_matching','distances'],
   install_requires = [
			'cleanco',
			'scikit-learn', 
                        'pandas',
                        'numpy',
			'tqdm'],
    long_description=long_description,
    long_description_content_type='text/markdown',
)



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tokenizer._whitespace.

Whitespace tokenizer
"""

from typing import Callable, Optional, Union

from ._regexp import RegexpTokenizer

__all__ = ['WhitespaceTokenizer']


class WhitespaceTokenizer(RegexpTokenizer):
    """A whitespace tokenizer.

    Examples
    --------
    >>> WhitespaceTokenizer().tokenize('a b c f a c g e a b')
    WhitespaceTokenizer({'a': 3, 'b': 2, 'c': 2, 'f': 1, 'g': 1, 'e': 1})


    .. versionadded:: 0.4.0

    """

    def __init__(
        self,
        scaler: Optional[Union[str, Callable[[float], float]]] = None,
        flags: int = 0,
    ) -> None:
        """Initialize tokenizer.

        Parameters
        ----------
        scaler : None, str, or function
            A scaling function for the Counter:

                - None : no scaling
                - 'set' : All non-zero values are set to 1.
                - 'length' : Each token has weight equal to its length.
                - 'length-log' : Each token has weight equal to the log of its
                   length + 1.
                - 'length-exp' : Each token has weight equal to e raised to its
                   length.
                - a callable function : The function is applied to each value
                  in the Counter. Some useful functions include math.exp,
                  math.log1p, math.sqrt, and indexes into interesting integer
                  sequences such as the Fibonacci sequence.
        flags : int
            Flags to pass to the regular expression matcher. See the
            `documentation on Python's re module
            <https://docs.python.org/3/library/re.html#re.A>`_ for details.


        .. versionadded:: 0.4.0

        """
        super(WhitespaceTokenizer, self).__init__(
            scaler, regexp=r'\S+', flags=flags
        )


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._damerau_levenshtein.

Damerau-Levenshtein distance
"""

from sys import maxsize
from typing import Any, Callable, List, Tuple, cast

from numpy import int_ as np_int
from numpy import zeros as np_zeros

from ._distance import _Distance

__all__ = [
    'DamerauLevenshtein',
]


class DamerauLevenshtein(_Distance):
    """Damerau-Levenshtein distance.

    This computes the Damerau-Levenshtein distance :cite:`Damerau:1964`.
    Damerau-Levenshtein code is based on Java code by Kevin L. Stern
    :cite:`Stern:2014`, under the MIT license:
    https://github.com/KevinStern/software-and-algorithms/blob/master/src/main/java/blogspot/software_and_algorithms/stern_library/string/DamerauLevenshteinAlgorithm.java
    """

    def __init__(
        self,
        cost: Tuple[float, float, float, float] = (1, 1, 1, 1),
        normalizer: Callable[[List[float]], float] = max,
        **kwargs: Any
    ):
        """Initialize Levenshtein instance.

        Parameters
        ----------
        cost : tuple
            A 4-tuple representing the cost of the four possible edits:
            inserts, deletes, substitutions, and transpositions, respectively
            (by default: (1, 1, 1, 1))
        normalizer : function
            A function that takes an list and computes a normalization term
            by which the edit distance is divided (max by default). Another
            good option is the sum function.
        **kwargs
            Arbitrary keyword arguments


        .. versionadded:: 0.4.0

        """
        super(DamerauLevenshtein, self).__init__(**kwargs)
        self._cost = cost
        self._normalizer = normalizer

    def dist_abs(self, src: str, tar: str) -> float:
        """Return the Damerau-Levenshtein distance between two strings.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        int (may return a float if cost has float values)
            The Damerau-Levenshtein distance between src & tar

        Raises
        ------
        ValueError
            Unsupported cost assignment; the cost of two transpositions must
            not be less than the cost of an insert plus a delete.

        Examples
        --------
        >>> cmp = DamerauLevenshtein()
        >>> cmp.dist_abs('cat', 'hat')
        1
        >>> cmp.dist_abs('Niall', 'Neil')
        3
        >>> cmp.dist_abs('aluminum', 'Catalan')
        7
        >>> cmp.dist_abs('ATCG', 'TAGC')
        2


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        ins_cost, del_cost, sub_cost, trans_cost = self._cost

        if src == tar:
            return 0
        if not src:
            return len(tar) * ins_cost
        if not tar:
            return len(src) * del_cost

        if 2 * trans_cost < ins_cost + del_cost:
            raise ValueError(
                'Unsupported cost assignment; the cost of two transpositions '
                + 'must not be less than the cost of an insert plus a delete.'
            )

        d_mat = np_zeros((len(src), len(tar)), dtype=np_int)

        if src[0] != tar[0]:
            d_mat[0, 0] = min(sub_cost, ins_cost + del_cost)

        src_index_by_character = {src[0]: 0}
        for i in range(1, len(src)):
            del_distance = d_mat[i - 1, 0] + del_cost
            ins_distance = (i + 1) * del_cost + ins_cost
            match_distance = i * del_cost + (
                0 if src[i] == tar[0] else sub_cost
            )
            d_mat[i, 0] = min(del_distance, ins_distance, match_distance)

        for j in range(1, len(tar)):
            del_distance = (j + 1) * ins_cost + del_cost
            ins_distance = d_mat[0, j - 1] + ins_cost
            match_distance = j * ins_cost + (
                0 if src[0] == tar[j] else sub_cost
            )
            d_mat[0, j] = min(del_distance, ins_distance, match_distance)

        for i in range(1, len(src)):
            max_src_letter_match_index = 0 if src[i] == tar[0] else -1
            for j in range(1, len(tar)):
                candidate_swap_index = (
                    -1
                    if tar[j] not in src_index_by_character
                    else src_index_by_character[tar[j]]
                )
                j_swap = max_src_letter_match_index
                del_distance = d_mat[i - 1, j] + del_cost
                ins_distance = d_mat[i, j - 1] + ins_cost
                match_distance = d_mat[i - 1, j - 1]
                if src[i] != tar[j]:
                    match_distance += sub_cost
                else:
                    max_src_letter_match_index = j

                if candidate_swap_index != -1 and j_swap != -1:
                    i_swap = candidate_swap_index

                    if i_swap == 0 and j_swap == 0:
                        pre_swap_cost = 0
                    else:
                        pre_swap_cost = d_mat[
                            max(0, i_swap - 1), max(0, j_swap - 1)
                        ]
                    swap_distance = (
                        pre_swap_cost
                        + (i - i_swap - 1) * del_cost
                        + (j - j_swap - 1) * ins_cost
                        + trans_cost
                    )
                else:
                    swap_distance = maxsize

                d_mat[i, j] = min(
                    del_distance, ins_distance, match_distance, swap_distance
                )
            src_index_by_character[src[i]] = i

        return cast(float, d_mat[len(src) - 1, len(tar) - 1])

    def dist(self, src: str, tar: str) -> float:
        """Return the Damerau-Levenshtein similarity of two strings.

        Damerau-Levenshtein distance normalized to the interval [0, 1].

        The Damerau-Levenshtein distance is normalized by dividing the
        Damerau-Levenshtein distance by the greater of
        the number of characters in src times the cost of a delete and
        the number of characters in tar times the cost of an insert.
        For the case in which all operations have :math:`cost = 1`, this is
        equivalent to the greater of the length of the two strings src & tar.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            The normalized Damerau-Levenshtein distance

        Examples
        --------
        >>> cmp = DamerauLevenshtein()
        >>> round(cmp.dist('cat', 'hat'), 12)
        0.333333333333
        >>> round(cmp.dist('Niall', 'Neil'), 12)
        0.6
        >>> cmp.dist('aluminum', 'Catalan')
        0.875
        >>> cmp.dist('ATCG', 'TAGC')
        0.5


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        if src == tar:
            return 0.0
        ins_cost, del_cost = self._cost[:2]
        return self.dist_abs(src, tar) / (
            self._normalizer([len(src) * del_cost, len(tar) * ins_cost])
        )


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._lig3.

LIG3 similarity
"""

from ._distance import _Distance
from ._levenshtein import Levenshtein

__all__ = ['LIG3']


class LIG3(_Distance):
    r"""LIG3 similarity.

    :cite:`Snae:2002` proposes three Levenshtein-ISG-Guth hybrid similarity
    measures: LIG1, LIG2, and LIG3. Of these, LIG1 is identical to ISG and LIG2
    is identical to normalized Levenshtein similarity. Only LIG3 is a novel
    measure, defined as:

        .. math::

            sim_{LIG3}(X, Y) = \frac{2I}{2I+C}

    Here, I is the number of exact matches between the two words, truncated to
    the length of the shorter word, and C is the Levenshtein distance between
    the two words.

    .. versionadded:: 0.4.1
    """

    _lev = Levenshtein()

    def sim(self, src: str, tar: str) -> float:
        """Return the LIG3 similarity of two words.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            The LIG3 similarity

        Examples
        --------
        >>> cmp = LIG3()
        >>> cmp.sim('cat', 'hat')
        0.8
        >>> cmp.sim('Niall', 'Neil')
        0.5714285714285714
        >>> cmp.sim('aluminum', 'Catalan')
        0.0
        >>> cmp.sim('ATCG', 'TAGC')
        0.0


        .. versionadded:: 0.4.1

        """
        if src == tar:
            return 1.0

        matches = 2 * sum(
            src[pos] == tar[pos] for pos in range(min(len(src), len(tar)))
        )
        cost = self._lev.dist_abs(src, tar)

        return matches / (matches + cost)


if __name__ == '__main__':
    import doctest

    doctest.testmod()
